///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef CAMERA_H
#define CAMERA_H


#include "TVecMat.h"
#include "Image.h"


struct Camera
{
	SqMat3double _rotation;
	Vec3double   _translation;
	SqMat3double _intrinsicMat;

	Vec2double _focalLength;
	Vec2double _principalPnt;
	double _skew;

	Vec<5, double> _distortion;

	void NormalizedCoord2PixelCoord(double &xp, double &yp, double xn, double yn) const 
	{
		xp = _focalLength[0] * ( xn + _skew * yn ) + _principalPnt[0];
		yp = _focalLength[1] * yn + _principalPnt[1];
	}

	void PixelCoord2NormalizedCoord(double &xn, double &yn, double xp, double yp) const
	{
		yn = ( yp - _principalPnt[1] ) / _focalLength[1];
		xn = ( xp - _principalPnt[0] ) / _focalLength[0] - _skew * yn;
	}

	void ApplyDistortion(double &xd, double &yd, double xn, double yn) const 
	{
		double xnyn = xn * yn;
		double xn2	= xn * xn;
		double yn2	= yn * yn;
		double r2	= xn2 + yn2;
		double s = 1.0 + r2 * ( _distortion[0] + r2 * ( _distortion[1] + r2 * _distortion[4] ) );
		double dx = 2 * _distortion[2] * xnyn + _distortion[3] * ( r2 + 2 * xn2 );
		double dy = _distortion[2] * ( r2 + 2 * yn2 ) + 2 * _distortion[3] * xnyn; 
		xd = s * xn + dx;
		yd = s * yn + dy;
	}

	Vec2double DistortedPixelCoord(const Vec2double &xp) const
	{
		double xn, yn;
		PixelCoord2NormalizedCoord(xn, yn, xp[0], xp[1]);
		double xd, yd;
		ApplyDistortion(xd, yd, xn, yn);
		Vec2double res;
		NormalizedCoord2PixelCoord(res[0], res[1], xd, yd);
		return res;
	}

	void xy2XY(double &X, double &Y, double x, double y, double Z) const
	{
		PixelCoord2NormalizedCoord(X, Y, x, y);
		X *= Z;
		Y *= Z;
	}

	void Parameter2Matrix(void);
	void Matrix2Parameter(void);

	void LoadFromFile(const char name[]);
	void SaveToFile(const char name[]);
};

istream &operator >> (istream &ifs, Camera &cam);
ostream &operator << (ostream &ofs, const Camera &cam);

#endif
