///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef HeapPQ_h
#define HeapPQ_h

#include <algorithm>
#include <valarray>
using std::swap;


template <typename V>
struct Heap1 //data and value are in one array;
{
	static inline int Left(int i) { return (i<<1)+1; }
	static inline int Right(int i) { return (i<<1)+2; }
	static inline int Parent(int i) { return (i-1)>>1; }

	static inline void RotateAt(V **varray, V *val, int i, const int N)
	{
		while (i<N)
		{
			int i_next = i; 
			const int left = Left(i);
			if (left < N)
			{
				if (val[i_next] < val[left]) i_next = left;
				
				const int right = left+1;
				if (right < N) 
				{
					if (val[i_next] < val[right]) i_next = right;					
				}
			}
			else break;

			if (i_next == i) break;
			else
			{
				swap(varray, val[i], val[i_next]);
				i = i_next;
			}
		}
	}

	static inline void BubbleUp(V **varray, V *val, int i)
	{
		while (i>0)
		{
			int i_parent = Parent(i);
			if (val[i_parent] < val[i])
			{
				swap(varray, val[i], val[i_parent]);
				i = i_parent;
			}
			else break;
		}
	}

	static inline void UpdateMax(V **varray, V *val, const int N)
	{
		RotateAt(varray, val, 0, N);
	}

	static inline void UpdateMax(V **varray, const V &v, V *val, const int N)
	{
		copy(varray, val[0], v);
		//dat[0] = d;
		RotateAt(varray, val, 0, N);
	}

	static inline void PopMax(V **varray, V *val, const int N)
	{
		copy(varray, val[0], val[N-1]);
		//dat[0] = dat[N-1];
		UpdateMax(varray, val, N-1);
	}

	static inline void Build(V **varray, V *val, const int N)
	{
		//starting from parent of N-1, rotate every one
		for (int i = Parent(N-1); i>=0; --i) RotateAt(varray, val, i, N);
	}

	static inline void Insert(V **varray, const V &v, V *val, const int N)
	{
		copy(varray, val[N], v);
		BubbleUp(varray, val, N);
	}

	static bool CheckHeap(const V *val, const int N)
	{
		//starting from N-1, rotate every one
		for (int i = N-1; i>0; --i) 
			if (val[Parent(i)]<val[i]) return false;
		return true;
	}

	static inline int MinI(const V *val, const int N)
	{
		int minI = 0;
		for (int i = 1; i < N; ++i) 
		{
			if (val[i] < val[minI]) minI = i;
		}
		return minI;
	}
	
	static inline void sort_ascending(V **varray, V *val, const int N)
	{
		Build(varray, val, N);
		for (int i = N-1; i > 0; --i)
		{
			swap(varray, val[i], val[0]);
			UpdateMax(varray, val, i);
		}
	}

	static inline void sort_descending(V **varray, V *val, const int N)
	{
		sort_ascending(varray, val, N);
		const int M = N/2;
		for (int i = 0; i < M; ++i)
		{
			int j = N-1-i; //i<j => 2*i<N-1
			swap(varray, val[i], val[j]);
		}
	}

};

#endif

