///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef Image_h
#define Image_h

#include <string>
#include <algorithm>
#include <typeinfo>
#include "RefCntMem.h"
#include "pixmap.h"

using std::string;
using std::swap;


class Image //row-major image using reference count memory
{
protected:

	int _nRows, _nCols, _nBands;		
	// the stride and gap are in the unit of pixel type, not byte!
	int _bandStride, _colStride, _rowStride;

	int _allBandStrides, _allColStrides, _allRowStrides;
	int _colGap, _rowGap;	
	int _nRowPixels, _nImgPixels;

	string _name;

protected:

	void ComputeGaps()
	{
		_allBandStrides   = _nBands*_bandStride;
		_allColStrides    = _nCols*_colStride;
		_allRowStrides    = _nRows*_rowStride;

		_colGap		= _colStride-_allBandStrides;
		_rowGap		= _rowStride-_allColStrides;
		
		_nRowPixels = _nCols*_nBands;
		_nImgPixels = _nRows*_nRowPixels;
	}

	void FreePixels()
	{
		_name.clear();
		_nRows = _nCols = _nBands = 0;
		_rowStride = _colStride = _bandStride = 0;
		ComputeGaps();
	}

public:

	Image() { FreePixels(); }

	bool Empty() const { return _nImgPixels == 0; }

	int Height() const { return _nRows; }
	int Width() const { return _nCols; }
	int NBands() const { return _nBands; }

	int BandStride() const { return _bandStride; }
	int ColStride() const { return _colStride; }
	int RowStride() const { return _rowStride; }

	int AllBandStrides() const { return _allBandStrides; }
	int AllColStrides() const { return _allColStrides; }
	int AllRowStrides() const { return _allRowStrides; }
	int ColGap() const { return _colGap; }
	int RowGap() const { return _rowGap; }	
	int NumRowPixels() const { return _nRowPixels; }
	int NumImgPixels() const { return _nImgPixels; }

	const string& Name() const { return _name; }
	string& Name() { return _name; }
};

inline double round(double x)
{ return floor(x+0.5); }

inline bool IsSameHeightAndWidth(const Image *I, const Image *J)
{ return I->Height() == J->Height() && I->Width() == J->Width(); }

inline bool IsSameSize(const Image *I, const Image *J)
{ return IsSameHeightAndWidth(I, J) && I->NBands() == J->NBands(); }

template <typename T>
class ImageOf : public Image
{
protected:

	RefCntMem<T> _pixels;    // reference counted memory
	T *_firstPixel;

protected:

	void MakeSubImage(int yO, int xO, int h, int w)
	{ 		
		_firstPixel = _firstPixel + yO * RowStride() + xO * ColStride(); 
		_nRows = h; _nCols = w; ComputeGaps();
	}

	void MakeBandImage(int b)
	{ 		
		_firstPixel = _firstPixel + b * BandStride(); 
		_nBands = 1; ComputeGaps(); 
	}

	void SubSampleImage(int ry, int rx)
	{
		_rowStride *= ry; _nRows = (_nRows-1)/ry+1; 
		_colStride *= rx; _nCols = (_nCols-1)/rx+1; 
		ComputeGaps();
	}

	void MakeTranspose(void)
	{ 
		swap(_colStride, _rowStride); 
		swap(_nCols, _nRows); ComputeGaps(); 
	}

public:
	ImageOf():Image() 
	{
		_firstPixel = 0;
		_pixels.ResetCount(0);
	}

	void FreePixels()
	{
		_firstPixel = 0;
		_pixels.ResetCount(0);
		Image::FreePixels();
	}

	void ResetPixels(T *pixels, 
		int height, int width, int nbands,
		int bandStride, int colStride, int rowStride)
	{
		_pixels.ResetCount(pixels);
		_firstPixel = pixels;
		
		_nBands = nbands;
		_nCols	= width;
		_nRows  = height;

		_bandStride = bandStride;
		_colStride  = colStride;
		_rowStride  = rowStride;

		ComputeGaps();
	}
	void ResetPixels(T *pixels, int height, int width, int nbands)
	{ 
		ResetPixels(pixels, height, width, nbands, 1, nbands, nbands*width); 
	}
	void ReAllocate(int height, int width, int nbands)
	{ 
		if (height!=Height()||width!=Width()||nbands!=NBands()) 
			ResetPixels(new T[height*width*nbands], height, width, nbands); 
	}

	bool IsMemContinuous() const 
	{ 
		return (_firstPixel && BandStride() == 1 && ColGap() == 0 && RowGap() == 0); 
	}

	T* FirstPixel() { return (T*)_firstPixel; }
	const T* FirstPixel() const { return (const T*)_firstPixel; }
	
	T* PixelAddress(int y, int x, int b) 
	{ return FirstPixel()+y*RowStride()+x*ColStride()+b*BandStride(); }
	const T* PixelAddress(int y, int x, int b) const 
	{ return FirstPixel()+y*RowStride()+x*ColStride()+b*BandStride(); }

	T& Pixel(int y, int x, int b) 
	{ return *(FirstPixel()+y*RowStride()+x*ColStride()+b*BandStride()); }
	const T& Pixel(int y, int x, int b) const 
	{ return *(FirstPixel()+y*RowStride()+x*ColStride()+b*BandStride()); }

	T* PixelAddress(int y, int x) 
	{ return FirstPixel()+y*RowStride()+x*ColStride(); }
	const T* PixelAddress(int y, int x) const 
	{ return FirstPixel()+y*RowStride()+x*ColStride(); }

	T& Pixel(int y, int x) 
	{ return *(FirstPixel()+y*RowStride()+x*ColStride()); }
	const T& Pixel(int y, int x) const 
	{ return *(FirstPixel()+y*RowStride()+x*ColStride()); }

	ImageOf<T> SubImage(int y, int x, int height, int width) const
	{ ImageOf<T> retval = *this; retval.MakeSubImage(y, x, height, width); return retval; }

	ImageOf<T> SubSampledImage(int ry, int rx) const
	{ ImageOf<T> retval = *this; retval.SubSampleImage(ry, rx); return retval; }

	ImageOf<T> BandImage(int band) const
	{ ImageOf<T> retval = *this; retval.MakeBandImage(band); return retval; }

	ImageOf<T> Transpose() const
	{ ImageOf<T> retval = *this; retval.MakeTranspose(); return retval; }

	inline void Fill(const T& a);

	bool InBounds(int y, int x, int b) const 
	{ 
		///return m_shape.InBounds(y, x, b);
		return (0 <= x && x < _nCols && 0 <= y && y < _nRows && 0 <= b && b < _nBands);
	}

	bool InBounds(int y, int x) const
	{ 
		///return m_shape.InBounds(y, x);
		return (0 <= x && x < _nCols && 0 <= y && y < _nRows);
	}  

	void SetPixel(T *p, const T &v)
	{
		///T *pLast = p + NBands() * BandSize();
		T *pLast = p + NBands() * BandStride();
		while (p != pLast)
		{
			*p = v;
			//NextBand(p);
			p += BandStride();
		}
	}

	void SetPixel(int y, int x, const T &v)
	{
		SetPixel(PixelAddress(y, x, 0), v);
	}

	void SetPixel(T *p, const T *v, int stride = 1)
	{
		///T *pLast = p + NBands() * BandSize();
		T *pLast = p + NBands() * BandStride();
		while (p != pLast)
		{
			*p = *v;
			//NextBand(p);
			p += BandStride();
			v += stride;
		}
	}	

	void SetPixel(int y, int x, const T *v, int stride = 1)
	{ 
		SetPixel(PixelAddress(y, x, 0), v, stride);
	}

	void GetPixel(const T *q, T *p, int interpMode, int stride = 1) const
	{ 
		const T *qLast = q + NBands() * BandStride();
		while (q != qLast)
		{
			*p = *q; 
			p += stride;
			q += BandStride();
		}
	}

};

template <typename D, typename S>
void CopyPixels(ImageOf<D>& dst, const ImageOf<S>& src);

template <typename T>
bool LoadPixels(const char *name, ImageOf<T> &img);

template <typename T>
bool SavePixels(const char *name, ImageOf<T> &img);

template <typename T>
bool WritePixels(const char *name, const ImageOf<T> &img);

template <typename T>
bool ReadPixels(const char *name, ImageOf<T> &img);

template <typename T>
void ImageOf<T>::Fill(const T& a)
{
	T *p = FirstPixel();
	const T *pR = p + AllRowStrides();

	const int AC = AllColStrides();
	const int AB = AllBandStrides();

	const int pbs = BandStride();
	const int pcg = ColGap();
	const int prg = RowGap();

	while (p != pR)
	{
		const T *pC = p + AC;
		while (p != pC)
		{
			const T *pB = p + AB;
			while (p != pB)
			{
				*p = a;
				p += pbs;
			}
			p += pcg;
		}
		p += prg;
	}
}

template <typename D, typename S>
void CopyPixels(ImageOf<D>& dst, const ImageOf<S>& src)
{
	dst.ReAllocate(src.Height(), src.Width(), src.NBands());

	const S* p = src.FirstPixel();
	D* q = dst.FirstPixel();
	const S *pR = p + src.AllRowStrides();

	const int AC = src.AllColStrides();
	const int AB = src.AllBandStrides();

	const int pbs = src.BandStride();
	const int qbs = dst.BandStride();
	const int pcg = src.ColGap();
	const int qcg = dst.ColGap();
	const int prg = src.RowGap();
	const int qrg = dst.RowGap();
	
	while (p != pR)
	{
		const S *pC = p + AC;
		while (p != pC)
		{
			const S *pB = p + AB;
			while (p != pB)
			{
				*q = (D)*p;
				p += pbs;
				q += qbs;
			}
			p += pcg;
			q += qcg;
		}
		p += prg;
		q += qrg;
	}
}

template <typename T>
bool WritePixels(const char *name, const ImageOf<T> &img)
{
	FILE *fp = fopen(name, "wb");

	if (fp==NULL) 
	{
		printf("WritePixels: fail to open %s\n", name);
		return false;
	}

	ImageOf<T> tmp;

	if (img.IsMemContinuous()) tmp = img;
	else CopyPixels(tmp, img);
	
	int W = img.Width();
	int H = img.Height();
	int B = img.NBands();

	int n;
	n = (int)fwrite(&H, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("WritePixels: fail to write height field in %s\n", name);
		return false;
	}

	n = (int)fwrite(&W, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("WritePixels: fail to write width field in %s\n", name);
		return false;
	}

	n = (int)fwrite(&B, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("WritePixels: fail to write band field in %s\n", name);
		return false;
	}

	n = (int)fwrite(tmp.FirstPixel(), sizeof(T), tmp.NumImgPixels(), fp);
	if (n != tmp.NumImgPixels())
	{
		printf("WritePixels: fail to write image data field in %s\n", name);
		return false;
	}
	
	fclose(fp);

	return true;
}

template <typename T>
bool ReadPixels(const char *name, ImageOf<T> &img)
{
	FILE *fp = fopen(name, "rb");

	if (fp==NULL) 
	{
		printf("ReadPixels: fail to open %s\n", name);
		return false;
	}

	int W, H, B;
	int n;

	n = (int)fread(&H, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("ReadPixels: fail to read height field in %s\n", name);
		return false;
	}

	n = (int)fread(&W, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("ReadPixels: fail to read width field in %s\n", name);
		return false;
	}

	n = (int)fread(&B, sizeof(int), 1, fp);
	if (n != 1)
	{
		printf("ReadPixels: fail to read band field in %s\n", name);
		return false;
	}

	if (!img.IsMemContinuous()) img.ReAllocate(H, W, B);

	n = (int)fread(img.FirstPixel(), sizeof(T), H*W*B, fp);
	if (n != H*W*B)
	{
		printf("ReadPixels: fail to read image data field in %s\n", name);
		return false;
	}

	fclose(fp);

	return true;

}

template <typename T>
bool LoadPixels(const char *name, ImageOf<T> &img)
{
	int L = (int)strlen(name);
	const char *ext = name+L-3;

	if (strcmp(ext,"PGM")==0 || strcmp(ext,"pgm")==0)
	{
		int width, height, nbands;
		unsigned char *pixels = 0;	
		int maxval;
		if (LoadPGM(name, pixels, width, height, maxval))
		{
			nbands = 1;
			if (maxval>255)
			{
				ImageOf<unsigned short> buf;
				buf.ResetPixels((unsigned short*)pixels, height, width, nbands);
				CopyPixels(img, buf);
			}
			else 
			{
				ImageOf<unsigned char> buf;
				buf.ResetPixels(pixels, height, width, nbands);
				CopyPixels(img, buf);
			}
		}
		else
		{
			printf("fail to load: %s\n", name);
			return false;
		}
	}
	else if (strcmp(ext,"PPM")==0 || strcmp(ext,"ppm")==0)
	{
		int width, height, nbands;
		unsigned char *pixels = 0;	
		int maxval;
		if (LoadPPM(name, pixels, width, height, maxval))
		{
			nbands = 3;
			if (maxval>255)
			{
				ImageOf<unsigned short> buf;
				buf.ResetPixels((unsigned short*)pixels, height, width, nbands);
				CopyPixels(img, buf);
			}
			else 
			{
				ImageOf<unsigned char> buf;
				buf.ResetPixels(pixels, height, width, nbands);
				CopyPixels(img, buf);
			}
		}
		else
		{
			printf("fail to load: %s\n", name);
			return false;
		}
	}
	else if (strcmp(ext,"PFM")==0 || strcmp(ext,"pfm")==0)
	{
		int width, height, nbands;
		float *pixels = 0;	
		float maxval;
		if (LoadPFM(name, pixels, width, height, maxval, nbands))
		{
			ImageOf<float> buf;
			buf.ResetPixels(pixels, height, width, nbands);
			CopyPixels(img, buf);
		}
		else
		{
			printf("fail to load: %s\n", name);
			return false;
		}
	}
	else
	{
		printf("unsupported format: %s\n", name);
		return false;
	}

	img.Name() = name;
	
	return true;
}

template <typename T>
bool SavePixels(const char *name, ImageOf<T> &img)
{
	int L = (int)strlen(name);
	const char *ext = name+L-3;

	int width = img.Width();
	int height = img.Height();
	int nbands = img.NBands();

	if (strcmp(ext,"PGM")==0 || strcmp(ext,"pgm")==0)
	{
		if (nbands != 1) 
		{
			printf("%s must have 1 color band!\n",name);
			return false;
		}
		int maxval;
		if (img.IsMemContinuous() && typeid(T)==typeid(unsigned char))
		{
			maxval = 255;
			unsigned char *pixels = (unsigned char*)img.FirstPixel();	
			SavePGM(name, pixels, width, height, maxval);	
		}
		else if (img.IsMemContinuous() && typeid(T)==typeid(unsigned short))
		{
			maxval = 65535;
			unsigned short *pixels = (unsigned short*)img.FirstPixel();				
			SavePGM(name, (unsigned char*)pixels, width, height, maxval);
		}
		else 
		{
			maxval = 255;
			ImageOf<unsigned char> buf;
			CopyPixels(buf, img);
			SavePGM(name, buf.FirstPixel(), width, height, maxval);	
		}
	}
	else if (strcmp(ext,"PPM")==0 || strcmp(ext,"ppm")==0)
	{
		if (nbands != 3) 
		{
			printf("%s must have 3 color band!\n",name);
			return false;
		}
		int maxval;
		if (img.IsMemContinuous() && typeid(T)==typeid(unsigned char))
		{
			maxval = 255;
			unsigned char *pixels = (unsigned char*)img.FirstPixel();
			SavePPM(name, pixels, width, height, maxval);	
		}
		else if (img.IsMemContinuous() && typeid(T)==typeid(unsigned short))
		{
			maxval = 65535;
			unsigned short *pixels = (unsigned short*)img.FirstPixel();	
			SavePPM(name, (unsigned char*)pixels, width, height, maxval);
		}
		else 
		{
			maxval = 255;
			ImageOf<unsigned char> buf;
			CopyPixels(buf, img);
			SavePPM(name, buf.FirstPixel(), width, height, maxval);	
		}
	}
	else if (strcmp(ext,"PFM")==0 || strcmp(ext,"pfm")==0)
	{
		float maxval = 1.0f;
		if (img.IsMemContinuous() && typeid(T)==typeid(float))
		{
			float *pixels = (float*)img.FirstPixel();
			SavePFM(name, pixels, width, height, maxval, nbands);	
		}
		else 
		{
			ImageOf<float> buf;
			CopyPixels(buf, img);
			SavePFM(name, buf.FirstPixel(), width, height, maxval, nbands);	
		}
	}
	else
	{
		printf("unsupported format: %s\n", name);
		return false;
	}
	return true;
}

template <typename T>
inline bool pixproc(ImageOf<T> &imq, ImageOf<T> &imp, T &a, void (*f)(T&, T&, T&a));

template <typename T> 
T imNormInf(ImageOf<T> &img);

template <typename T>
T imNorm1(ImageOf<T> &img);

template <typename T>
T imNorm2(ImageOf<T> &img);

template <typename T>
bool imInnerProd(T &a, ImageOf<T> &im0, ImageOf<T> &im1);

template <typename T>
ImageOf<T> & operator += (ImageOf<T> &imq, ImageOf<T> &imp);

template <typename T>
ImageOf<T> & operator -= (ImageOf<T> &imq, ImageOf<T> &imp);

template <typename T>
ImageOf<T> & operator *= (ImageOf<T> &imq, ImageOf<T> &imp);

template <typename T>
ImageOf<T> & operator /= (ImageOf<T> &imq, ImageOf<T> &imp);

template <typename T>
ImageOf<T> & operator += (ImageOf<T> &img, T &a);

template <typename T>
ImageOf<T> & operator -= (ImageOf<T> &img, T &a);

template <typename T>
ImageOf<T> & operator *= (ImageOf<T> &img, T &a);

template <typename T>
ImageOf<T> & operator /= (ImageOf<T> &img, T &a);

template <typename T>
bool imMoveAssign(ImageOf<T> &dst, ImageOf<T> &src, T &a);

template <typename T>
bool imFilter(ImageOf<T> &dst, ImageOf<T> &src, const double *knl, int kry, int krx);

template <typename T, typename PROC>
bool pixproc(ImageOf<T> &imq, ImageOf<T> &imp, T &a, PROC &proc)
{
	if (!IsSameSize(&imq,&imp)) return false;

	T* p = imp.FirstPixel();
	T* q = imq.FirstPixel();
    const T *pR = p + imp.AllRowStrides();

	const int AC = imp.AllColStrides();
	const int AB = imp.AllBandStrides();

	const int pbs = imp.BandStride();
	const int qbs = imq.BandStride();
	const int pcg = imp.ColGap();
	const int qcg = imq.ColGap();
	const int prg = imp.RowGap();
	const int qrg = imq.RowGap();

	while (p != pR)
	{
		const T *pC = p + AC;
		while (p != pC)
		{
			const T *pB = p + AB;
			while (p != pB)
			{
				PROC::func(*q, *p, a);
				p += pbs;
				q += qbs;
			}
			p += pcg;
			q += qcg;
		}
		p += prg;
		q += qrg;
	}

	return true;
}

template <typename T>
struct pxAbs { public: static inline void func(T& y, T& x, T& a) { y = (x>0)?x:-x; } };

template <typename T>
struct pxNormInf { public: static inline void func(T& y, T& x, T& a) { a = max(a, abs(y)); } };

template <typename T>
struct pxNorm1 { public: static inline void func(T& y, T& x, T& a) { a += abs(y); } };

template <typename T>
struct pxNorm2 { public: static inline void func(T& y, T& x, T& a) { a += y*y; } };

template <typename T>
struct pxInnerProd { public: static inline void func(T& y, T& x, T& a) { a += y*x; } };

template <typename T>
struct pxPlusAssignS { public: static inline void func(T& y, T& x, T& a) { y += a; } };

template <typename T>
struct pxMinusAssignS { public: static inline void func(T& y, T& x, T& a) { y -= a; } };

template <typename T>
struct pxTimesAssignS { public: static inline void func(T& y, T& x, T& a) { y *= a; } };

template <typename T>
struct pxOverAssignS { public: static inline void func(T& y, T& x, T& a) { y /= a; } };

template <typename T>
struct pxPlusAssign { public: static inline void func(T& y, T& x, T& a) { y += x; } };

template <typename T>
struct pxMinusAssign { public: static inline void func(T& y, T& x, T& a) { y -= x; } };

template <typename T>
struct pxTimesAssign { public: static inline void func(T& y, T& x, T& a) { y *= x; } };

template <typename T>
struct pxOverAssign { public: static inline void func(T& y, T& x, T& a) { y /= x; } };

template <typename T>
struct pxMoveAssign { public: static inline void func(T& y, T& x, T& a) { y += a*x; } };

template <typename T>
void imAbs (ImageOf<T> &dst, ImageOf<T> &src)
{
	dst.ReAllocate(src.Height(), src.Width(), src.NBands());

	T a = T(0);
	pixproc(dst, src, a, pxAbs<T>());
}

template <typename T> 
T imNormInf(ImageOf<T> &img)
{
	T a = T(0);
	pixproc(img, img, a, pxNormInf<T>());
	return a;
}

template <typename T> 
T imNorm1(ImageOf<T> &img)
{
	T a = T(0);
	pixproc(img, img, a, pxNorm1<T>());
	return a/(T)img.NumImgPixels();
}

template <typename T> 
T imNorm2(ImageOf<T> &img)
{
	T a = T(0);
	pixproc(img, img, a, pxNorm2<T>());
	return (T)sqrt(a/(T)img.NumImgPixels());
}

template <typename T>
bool imInnerProd(T &a, ImageOf<T> &im0, ImageOf<T> &im1)
{
	if (!IsSameSize(&im0, &im1)) return false; 

	a = T(0);
	pixproc(im0, im1, a, pxInnerProd<T>());
	return true;
}

template <typename T>
ImageOf<T> & operator += (ImageOf<T> &img, T &a)
{
	pixproc(img, img, a, pxPlusAssignS<T>());
	return img;
}

template <typename T>
ImageOf<T> & operator -= (ImageOf<T> &img, T &a)
{
	pixproc(img, img, a, pxMinusAssignS<T>());
	return img;
}

template <typename T>
ImageOf<T> & operator *= (ImageOf<T> &img, T &g)
{
	T a = T(g);
	pixproc(img, img, a, pxTimesAssignS<T>());
	return img;
}

template <typename T>
ImageOf<T> & operator /= (ImageOf<T> &img, T &g)
{
	T a = T(g);
	pixproc(img, img, a, pxOverAssignS<T>());
	return img;
}

template <typename T>
ImageOf<T> & operator += (ImageOf<T> &imq, ImageOf<T> &imp)
{
	T a = T(0);
	if (IsSameSize(&imq, &imp)) pixproc(imq, imp, a, pxPlusAssign<T>());
	return imq;
}

template <typename T>
ImageOf<T> & operator -= (ImageOf<T> &imq, ImageOf<T> &imp)
{
	T a = T(0);
	if (IsSameSize(&imq, &imp)) pixproc(imq, imp, a, pxMinusAssign<T>());
	return imq;
}

template <typename T>
ImageOf<T> & operator *= (ImageOf<T> &imq, ImageOf<T> &imp)
{
	T a = T(0);
	if (IsSameSize(&imq, &imp)) pixproc(imq, imp, a, pxTimesAssign<T>());
	return imq;
}

template <typename T>
ImageOf<T> & operator /= (ImageOf<T> &imq, ImageOf<T> &imp)
{
	T a = T(0);
	if (IsSameSize(&imq, &imp)) pixproc(imq, imp, a, pxOverAssign<T>());
	return imq;
}

template <typename T>
bool imMoveAssign(ImageOf<T> &dst, ImageOf<T> &src, T &a)
{
	if (!IsSameSize(&dst, &src)) return false;

	pixproc(dst, src, a, pxMoveAssign<T>());
	return true;
}

template <typename T>
bool imFilter(ImageOf<T> &dst, ImageOf<T> &src, const double *knl, int kry, int krx)
//knl is assumed to be in row major
{
	dst.ReAllocate(src.Height(), src.Width(), src.NBands());

	dst.Fill(T(0));

	int ksy = 2*kry+1;
	int ksx = 2*krx+1;
	int H = src.Height()-ksy+1;
	int W = src.Width()-ksx+1;

	ImageOf<T> dst_cnt = dst.SubImage(kry, krx, H, W);	

	for (int y = 0; y < ksy; y++)
	{
		for (int x = 0; x < ksx; x++)
		{
			double a = knl[y*ksx+x];

			ImageOf<T> src_cnt = src.SubImage(y, x, H, W);

			if (!imMoveAssign(dst_cnt, src_cnt, a)) return false;
		}
	}
	return true;
}

inline double ResampleLinear(double v0, double v1, double f)
{
    return v0 + f * (v1 - v0);
}

inline double ResampleCubicV(double v0, double v1, double v2, double v3, double f)
{
	// 0.5 * [f^3 f^2 f^1 f^0] * [-1 3 -3 1; 2 -5 4 -1; -1 0 1 0; 0 2 0 0] * [v0; v1; v2; v3]
	double v0_2 = v0+v0;

	double v1_2 = v1+v1;
	double v1_3 = v1_2+v1;
	double v1_5 = v1_2+v1_3;
	
	double v2_2 = v2+v2;
	double v2_3 = v2_2+v2;
	double v2_4 = v2_3+v2;

    double a0 = (v1_3+v3)-(v0+v2_3);
	double a1 = (v0_2+v2_4)-(v1_5+v3);
	double a2 = v2-v0;
	double a3 = v1_2;

	double v = 0.5*(((a0*f+a1)*f+a2)*f+a3);
	return v;
}

inline double ResampleCubicD(double v0, double v1, double v2, double v3, double f)
{
	// 0.5 * [f^3 f^2 f^1 f^0] * [-1 3 -3 1; 2 -5 4 -1; -1 0 1 0; 0 2 0 0] * [v0; v1; v2; v3]
	double v0_2 = v0+v0;

	double v1_2 = v1+v1;
	double v1_3 = v1_2+v1;
	double v1_5 = v1_2+v1_3;
	
	double v2_2 = v2+v2;
	double v2_3 = v2_2+v2;
	double v2_4 = v2_3+v2;

    double a0 = (v1_3+v3)-(v0+v2_3);
	double a1 = (v0_2+v2_4)-(v1_5+v3);
	double a2 = v2-v0;
	double a3 = v1_2;

	// v = 0.5*(((a0*f+a1)*f+a2)*f+a3);
	// d = 0.5*(((a0*f+a1)*f+a2)+((a0*f+a1)+(a0)*f)*f) 
	//	 = 0.5*(3*a0*f^2 + 2*a1*f + a2)
	return 0.5 * ( ( 3 * a0 * f + 2 * a1 ) * f + a2 );
}

inline void ResampleCubicVD(double &v, double &d, double v0, double v1, double v2, double v3, double f)
{
	// 0.5 * [f^3 f^2 f^1 f^0] * [-1 3 -3 1; 2 -5 4 -1; -1 0 1 0; 0 2 0 0] * [v0; v1; v2; v3]
	double v0_2 = v0+v0;

	double v1_2 = v1+v1;
	double v1_3 = v1_2+v1;
	double v1_5 = v1_2+v1_3;
	
	double v2_2 = v2+v2;
	double v2_3 = v2_2+v2;
	double v2_4 = v2_3+v2;

    double a0 = (v1_3+v3)-(v0+v2_3);
	double a1 = (v0_2+v2_4)-(v1_5+v3);
	double a2 = v2-v0;
	double a3 = v1_2;

	// v = 0.5*(((a0*f+a1)*f+a2)*f+a3);
	// d = 0.5*(((a0*f+a1)*f+a2)+((a0*f+a1)+(a0)*f)*f) 
	//	 = 0.5*(3*a0*f^2 + 2*a1*f + a2)
	double a0f		= a0 * f;
	double a0f1		= a0f + a1;
	double a0f1f	= a0f1 * f;
	double a0f1f2	= a0f1f + a2;
	double a0f1f2f	= a0f1f2 * f;
	double a0f1f2f3 = a0f1f2f + a3;

	v = 0.5 * a0f1f2f3;
	d = 0.5 * (a0f1f2 + (a0f1 + a0f)*f);	
}

template <class T>
inline T ResampleBiLinear(T *src, int oH, int oV, double xf, double yf)
{
    // Resample a pixel using bilinear interpolation
    double h1 = ResampleLinear(src[0 ], src[   oH], xf);
    double h2 = ResampleLinear(src[oV], src[oV+oH], xf);
    double  v = ResampleLinear(h1, h2, yf);
    return (T)v;
} 

template <class T>
inline T ResampleBiCubic(T *src, int oH, int oV, double xf, double yf)
{
    // Resample a pixel using bilinear interpolation
    double h[4];
    for (int i = 0; i < 4; i++)
    {
        int j = (i-1)*oV;
		h[i] = ResampleCubicV(src[j-oH], src[j], src[j+oH], src[j+2*oH], xf);
    }
    double  v = ResampleCubicV(h[0], h[1], h[2], h[3], yf);

    return (T)v;
}

#endif 



