///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef Ref_Cnt_Mem_h
#define Ref_Cnt_Mem_h


template <typename T>
class RefCntMem            // reference-counted memory object class
{

//	friend class RefCntMem;

private:

	struct CountedPtr			// counted pointer
	{
		T *_data;				// allocated data
		int _count;				// reference count
		CountedPtr() { _data = 0; _count = 0; }
	};
	CountedPtr *_cptr;

    void DecrementCount(void)  // decrement the reference count and delete if done
	{		
		if (_cptr)
		{
			_cptr->_count--;
			if (_cptr->_count == 0)
			{
				delete[] _cptr->_data; 
				delete _cptr;
				_cptr = 0;
			}
		}
	}
    void IncrementCount(void)  // increment the reference count
	{
		if (_cptr) _cptr->_count++;
	}
    
public:
	RefCntMem(void)                   // default constructor
	{ 
		_cptr = 0; 
	}

	RefCntMem(size_t n)
	{
		_cptr = 0;
		ReAllocate(n);
	}

    RefCntMem(const RefCntMem <T>& rcm)   // copy constructor
	{	
		_cptr = 0;    
		(*this) = rcm;  
	}

    ~RefCntMem(void)				  // destructor
	{
		DecrementCount();
		_cptr = 0;      
	}

    RefCntMem& operator=(const RefCntMem<T>& rcm)  // assignment
	{
		DecrementCount();   
		_cptr = rcm._cptr;
		IncrementCount();
		return *this;
	}

    void ResetCount(T *data)        // start count new memory
	{
		DecrementCount();
		if (data)
		{
			_cptr = new CountedPtr;
			_cptr->_data = data;
			_cptr->_count = 1;			
		}
		else
			_cptr = 0; 		
	}

	void ReAllocate(size_t n)
	{
		ResetCount(new T [n]);
	}

	T* Memory(void)							// pointer to allocated memory
	{
		return (_cptr) ? _cptr->_data : 0;
	}
	
	operator const T* () const 
	{
		return Memory();
	}

	operator T* ()  
	{
		return Memory();
	}
};

#endif // Ref_Cnt_Mem_h


