///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef TVec_inl
#define TVec_inl


template <int N, class floating> 
Vec <N,floating> Vec<N,floating>::zero(void)
{
	Vec <N,floating> z;
	int d;
	for (d=0;d<N;d++)
	{
		z[d] = (floating)0;
	}
	return z;
}

template <int N, class floating> 
Vec <N,floating> Vec<N,floating>::unit(int i)
{
	Vec <N,floating> u = Vec<N,floating>::zero();
	u[i] = (floating)1;
	return u;
}

template <int N, class floating> 
Vec <N,floating> ::Vec( const floating vec[N])
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] = vec[d];
	}
}

template <int N, class floating> 
Vec <N,floating> ::Vec( const Vec <N,floating> &vec)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] = vec[d];
	}
}

template <int N, class floating>
bool Vec <N,floating> ::isconstant(floating f) const
{
	int d;
	for (d=0;d<N;d++)
	{
		if (v[d] != f)
		{
			return false;
		}
	}
	return true;
}

template <int N, class floating>
bool Vec <N,floating> ::iszero(void) const
{
	int d;
	for (d=0;d<N;d++)
	{
		if (v[d] != (floating)0)
		{
			return false;
		}
	}
	return true;
}

template <int N, class floating>
bool Vec<N,floating>::operator == (const Vec <N, floating> &vec) const
{
	int d;
	for (d=0;d<N;d++)
	{
		if (v[d] != vec[d])
		{
			return false;
		}
	}
	return true;
}

template <int N, class floating>
void Vec<N,floating>::makezero(void)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] = (floating)0;
	}
}

template <int N, class floating>
void Vec<N,floating>::makeunit(int i)
{
	makezero();
	v[i] = (floating)1;
}

template <int N, class floating> 
Vec <N, floating> &Vec<N,floating>::operator = (const Vec <N, floating> &vec)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] = vec[d];
	}
	return *this;
}

template <int N, class floating> 
Vec <N, floating> &Vec<N,floating>::operator = (const floating &f)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] = f;
	}
	return *this;
}

template <int N, class floating> 
Vec <N, floating>& Vec<N,floating>::operator += (const Vec <N, floating> &vec)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] += vec[d];
	}
	return *this;
}

template <int N, class floating> 
Vec <N, floating> &Vec<N,floating>::operator -= (const Vec <N, floating> &vec)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] -= vec[d];
	}
	return *this;
}

template <int N, class floating> 
Vec <N, floating> &Vec<N,floating>::operator *= (floating s)
{
	int d;
	for (d=0;d<N;d++)
	{
		v[d] *= s;
	}
	return *this;
}

template <int N, class floating>
Vec <N, floating> operator - (const Vec <N, floating>&v)
{
	Vec <N, floating> n;
	int d;
	for (d=0;d<N;d++)
	{
		n[d]=-v[d];
	}
	return n;
}

template <int N, class floating>
floating inner(const Vec <N,floating> &a, const Vec <N,floating> &b)
{
	floating sum=0;
	int d;
	for (d=0;d<N;d++)
	{
		sum+=a[d]*b[d];
	}
	return sum;
}

template <class floating> 
Vec<3, floating> cross(const Vec<3,floating> &a, const Vec<3,floating> &b)
{
	Vec<3, floating> v;
	v[0] = a[1] * b[2] - a[2] * b[1];
	v[1] = a[2] * b[0] - a[0] * b[2];
	v[2] = a[0] * b[1] - a[1] * b[0];
	return v;
}

template <int N, class floating>
istream& operator >> (istream &is, Vec<N, floating> &m)
{
	int j;

	for (j=0;j<N;j++)
	{
		is>>m[j];
	}
	return is;
}

template <int N, class floating>
ostream& operator << (ostream &os, const Vec<N, floating> &m)
{
	int j;

	for (j=0;j<N;j++)
	{
		os<<m[j]<<"\t";
	}
	//os<<endl;
	return os;
}

#endif
