///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith Li Zhang                              //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "data.h"

void Parameters::Parse(
	int argc, 
	const char *argv[])
{
	// For an explanation of the following parameters, please read the 
	// readme.txt file or the comments in data.h.

	vector<string> allArgv;
	collectAllArgs(allArgv, argc, argv);
	bool good;


	/////////////////////////// REQUIRED PARAMETERS ///////////////////////////

	good = extractUnknownParaFromCommandArg(
		_imFiles, 
		"image files (imFiles)", 
		"-imFiles", 
		allArgv);
	if (!good)
		exit(0);

	const int numCams = (int)_imFiles.size();
	if (numCams < 2)
	{
		printf("ERROR: mvstereo requires at least two input images!\n");
		exit(0);
	}

	good = extractUnknownParaFromCommandArg(
		_camFiles, 
		"camera files (camFiles)", 
		"-camFiles", 
		allArgv);
	if (!good) 
		exit(0);

	if (_camFiles.size() != numCams)
	{
		printf("ERROR: number of cameras is different from num of images!\n");
		exit(0);
	}
	
	good = extractUnknownParaFromCommandArg(
		_camPairs, 
		"camera pairs (camPairs)", 
		"-camPairs", 
		allArgv);

	if (!good) 
		exit(0);

	if ((int)_camPairs.size() < numCams-1)
	{
		printf("ERROR: number of camera pairs should be at least %d!\n", numCams-1);
		exit(0); 
	}

	// check to make sure the camera pairs correspond to images
	for (int i = 0; i < numCams; ++i)
	{
		int found = 0;
		for (int p = 0; p < (int)_camPairs.size(); ++p)
		{
			if (i==_camPairs[p][0] || i==_camPairs[p][1])
			{
				found = 1;
				break;
			}
		}
		if (!found)
		{
			printf("ERROR: image %d is not being matched!\n", i);
			exit(0);
		}
	}

	good = extractUnknownParaFromCommandArg(
		_zMinInv, 
		"inverse of the min depth (zMinInv)", 
		"-zMinInv", 
		allArgv);

	if (!good) 
		exit(0);


	/////////////////////////// OPTIONAL PARAMETERS ///////////////////////////

	_zMaxInv = 0.0;
	extractDefaultParaFromCommandArg(
		_zMaxInv, 
		"inverse of the max depth (zMaxInv)", 
		"-zMaxInv", 
		allArgv);

	if (_zMaxInv >= _zMinInv)
	{
		printf("ERROR: invalid input : zMaxInv >= zMinInv!\n");
		exit(0);
	}

	_numLayers = 32;
	extractDefaultParaFromCommandArg(
		_numLayers, 
		"number of layers (numLayers)", 
		"-numLayers", 
		allArgv);

	if (_numLayers < 1)
	{ 
		printf("ERROR: should have at least 1 disparity label!\n");
		exit(0);
	}

	_lbl2invDepth.resize(_numLayers);
	for (int i = 0; i < _numLayers; ++i)
	{
		double iz;
		if(_numLayers < 2)
		{
			iz = 0;
		}
		else
		{
			iz = double(i)/double(_numLayers-1)*(_zMinInv-_zMaxInv)+_zMaxInv;
		}
		_lbl2invDepth[i] = iz;
	}

	_depthScale = 8;
	extractDefaultParaFromCommandArg(
		_depthScale, 
		"disparity/depth scale (dptScale)", 
		"-dptScale", 
		allArgv);

	if (_depthScale<1)
	{
		printf("ERROR: depth scale should be at least 1!\n");
		exit(0);
	}

	_dptFiles.resize(numCams);
	extractDefaultParaFromCommandArg(
		_dptFiles, 
		"disparity/depth files (dptFiles)", 
		"-dptFiles", 
		allArgv);

	if (_dptFiles.size() != numCams)
	{
		printf("ERROR: number of disparity/depth output files is different from number images!\n");
		exit(0);
	}

	_dptSmoothFiles.resize(numCams);
	extractDefaultParaFromCommandArg(
		_dptSmoothFiles, 
		"filtered/smooth depth files (dptSmoothFiles)", 
		"-dptSmoothFiles", 
		allArgv);

	_initDptValue = 0;
	extractDefaultParaFromCommandArg(
		_initDptValue, 
		"initial disparity/depth value (initDptValue)", 
		"-initDptValue", 
		allArgv);

	if (_initDptValue<0 || _initDptValue>_numLayers)
	{
		printf("ERROR: init depth value out of range!\n");
		exit(0);
	}

	_birchfield = 1;
	extractDefaultParaFromCommandArg(
		_birchfield, 
		"Birchfield-Tomasi (birchfield)", 
		"-birchfield", 
		allArgv);
	
	// 17.5 generates good results for two view stereo, ~8 for five view stereo
	_lambda = 35.0f / (float)numCams; 
	extractDefaultParaFromCommandArg(
		_lambda, 
		"disparity spatial smoothness weight (lambda)", 
		"-lambda", 
		allArgv);

	_tauSm = 2.0f;
	extractDefaultParaFromCommandArg(
		_tauSm, 
		"disparity spatial smoothness robustness (tauSm)", 
		"-tauSm", 
		allArgv);

	_tauPh = 30.0f;
	extractDefaultParaFromCommandArg(
		_tauPh, 
		"disparity-mapped photo-conisistency robustness (tauPh)", 
		"-tauPh", 
		allArgv);

	_lambdaVs = 100000.0f;
	extractDefaultParaFromCommandArg(
		_lambdaVs, 
		"disparity visiblity violation penalty (lambdaVs)", 
		"-lambdaVs", 
		allArgv);

	_sigmaX = 20.0f;
	extractDefaultParaFromCommandArg(
		_sigmaX, 
		"scale of geometric neighborhoods (sigmaX)", 
		"-sigmaX", 
		allArgv);

	if(_sigmaX <= 0.0f)
	{
		printf("scale of geometric neighborhoods must be positive!\n");
		exit(0);
	}

	_sigmaC = 4.0f;
	extractDefaultParaFromCommandArg(
		_sigmaC, 
		"scale of chromatic neighborhoods (sigmaC)", 
		"-sigmaC", 
		allArgv);

	if(_sigmaC <= 0.0f)
	{
		printf("ERROR: scale of chromatic neighborhoods must be positive!\n");
		exit(0);
	}

	_sigmaD = 2.0f;
	extractDefaultParaFromCommandArg(
		_sigmaD, 
		"scale of disparity/depth label bandwidth (sigmaD)", 
		"-sigmaD", 
		allArgv);

	if(_sigmaD <= 0.0f)
	{
		printf("ERROR: scale of depth label bandwidth must be positive!\n");
		exit(0);
	}

	_dataCostType = L2; // L2 = 1
	extractDefaultParaFromCommandArg(
		_dataCostType, 
		"data cost type (dataCostType)", 
		"-dataCostType", 
		allArgv);

	if (_dataCostType != L1 && _dataCostType != L2)
	{
		printf("ERROR: data cost type needs to be either L1 (0) or L2 (1)!\n");
		exit(0);
	}

	_maxIter = 2;
	extractDefaultParaFromCommandArg(
		_maxIter, 
		"maximum number of iterations (maxIter)", 
		"-maxIter", 
		allArgv);

	_nbrEdgeFiles.resize(numCams);
	extractDefaultParaFromCommandArg(
		_nbrEdgeFiles, 
		"neighborhood MST edge set (nbrEdgeFiles)", 
		"-nbrEdgeFiles", 
		allArgv);

	_runMstnbr = 1;
	extractDefaultParaFromCommandArg(
		_runMstnbr, 
		"run minimum spanning tree neighborhood computation (runMstnbr)", 
		"-runMstnbr", 
		allArgv);

	_runMvstereo = 1;
	extractDefaultParaFromCommandArg(
		_runMvstereo, 
		"run multiple-view stereo computation (runMvstereo)", 
		"-runMvstereo", 
		allArgv);

	_runTrilatfilt = 1;
	extractDefaultParaFromCommandArg(
		_runTrilatfilt, 
		"run trilateral filtering (runTrilatfilt)", 
		"-runTrilatfilt", 
		allArgv);

	_runNoiseremoval = 1;
	extractDefaultParaFromCommandArg(
		_runNoiseremoval, 
		"run noise removal (runNoiseremoval)", 
		"-runNoiseremoval", 
		allArgv);

	_nrThresh = 3;
	extractDefaultParaFromCommandArg(
		_nrThresh, 
		"noise removal threshold (nrThresh)", 
		"-nrThresh", 
		allArgv);

	_maxNbrRad = 40;
	extractDefaultParaFromCommandArg(
		_maxNbrRad, 
		"maximum neighborhood radius (maxNbrRad)", 
		"-maxNbrRad", 
		allArgv);

	_numMst = 2;
	extractDefaultParaFromCommandArg(
		_numMst, 
		"number of minimum spanning trees (numMst)", 
		"-numMst", 
		allArgv);

	_numThreads = numCams;
	extractDefaultParaFromCommandArg(
		_numThreads, 
		"number of threads (numThreads)", 
		"-numThreads", 
		allArgv);

	//construct vector of labels
	_labels.resize(_numLayers);
	for(int d = 0; d < _numLayers; d++)
	{
		_labels[d] = _numLayers - 1 - d;
	}
}


