///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef DATA_H
#define DATA_H

#include <valarray>
#include "CmdArgParser.h"
#include "Camera.h"
#include "Image.h"
#include "energy.h"


struct Coord
{
	int x, y;
	Coord() {}
	Coord(int a, int b) { x = a; y = b; }
	Coord operator- ()        { return Coord(-x, -y); }
	Coord operator+ (Coord a) { return Coord(x + a.x, y + a.y); }
	Coord operator- (Coord a) { return Coord(x - a.x, y - a.y); }
	bool  operator< (Coord a) { return (x <  a.x) && (y <  a.y); }
	bool  operator<=(Coord a) { return (x <= a.x) && (y <= a.y); }
	bool  operator> (Coord a) { return (x >  a.x) && (y >  a.y); }
	bool  operator>=(Coord a) { return (x >= a.x) && (y >= a.y); }
	bool  operator==(Coord a) { return (x == a.x) && (y == a.y); }
	bool  operator!=(Coord a) { return (x != a.x) || (y != a.y); }
};

struct Coordf
{
	float x, y;
	Coordf() {}
	Coordf(float a, float b) { x = a; y = b; }
	Coordf operator- ()        { return Coordf(-x, -y); }
	Coordf operator+ (Coordf a) { return Coordf(x + a.x, y + a.y); }
	Coordf operator- (Coordf a) { return Coordf(x - a.x, y - a.y); }
	bool  operator< (Coordf a) { return (x <  a.x) && (y <  a.y); }
	bool  operator<=(Coordf a) { return (x <= a.x) && (y <= a.y); }
	bool  operator> (Coordf a) { return (x >  a.x) && (y >  a.y); }
	bool  operator>=(Coordf a) { return (x >= a.x) && (y >= a.y); }
	bool  operator==(Coordf a) { return (abs(x - a.x) < 0.0001f) && (abs(y - a.y) < 0.0001f); }
	bool  operator!=(Coordf a) { return (abs(x - a.x) > 0.0001f) || (abs(y - a.y) > 0.0001f); }
};

struct DataIn
{
	// array of camera parameters
	valarray< Camera >                   _camArray;
	// array of input images
	valarray< ImageOf<short> >           _imArray;
	// array of input image pixel bounds
	valarray< ImageOf<short> >           _imArrayBound;
	// array of degree functions
	valarray< ImageOf<float> >           _degFuncArray; 
	// array of neighborhood pixel edge sets
	valarray< ImageOf< vector<Coord> > > _nbrEdgeArray;
};

struct DataOut
{
	// array of output disparity/depth maps (stored as labels)
	valarray< ImageOf<short> >           _lblArray;
	// array of subpixel accuracy (smooth) disparity maps
	valarray< ImageOf<double> >          _dptSmoothArray;
	// array of variable arrays used in the graph cuts code
	valarray< ImageOf<Energy::Var> >     _varArray;
};

struct Parameters
{
	/////////////////////////// REQUIRED PARAMETERS ///////////////////////////

	// vector of input image filename strings
	vector<string>   _imFiles; // required parameter

	// vector of camera parameter filename strings
	vector<string>   _camFiles; // required parameter

	// vector of camera pairs, given as indices starting at 0
	vector<Vec2int>  _camPairs; // required parameter

	// the inverse of the minimum depth 
	double           _zMinInv; // required parameter


	/////////////////////////// OPTIONAL PARAMETERS ///////////////////////////
	//  optional only in the sense that the algorithm will run and it will   //
	//  generate reasonable results, but you will rarely achieve the best    //
	//  results if you leave these parameters to their default settings      //
    ///////////////////////////////////////////////////////////////////////////

	// run (=1) or do not run (=0) the minimum spanning tree computation. if
	// _runMstnbr = 0 it is assumed that the neighborhood structure has
	// been precomputed
	int              _runMstnbr; // default = 1

	// run (=1) or do not run (=0) the multiple-view stereo algorithm
	int              _runMvstereo; // default = 1

	// run (=1) or do not run (=0) the trilateral filter post-processing
	int              _runTrilatfilt; // default = 1

	// run (=1) or do not run (=0) disparity/depth map noise removal
	int              _runNoiseremoval; // default = 1

	// the number of threads allowed for computing the minimum spanning
	// tree neighborhoods and for the trilateral filter post-process. a
	// separate thread will be giving to each image since the processing
	// can be done in parallel.
	int              _numThreads; // default = <number of images>

	// use Birchfield-Tomasi technique (=1) or don't (=0)
	int              _birchfield; // default = 1

	// the type of data cost, L1 (0) or L2 (1)
	int              _dataCostType; // default = 1 (L2)
	enum { L1, L2 };

	// the regularization/smoothness term coefficient
	float            _lambda; // default = 35 / numCams

	// disparity visibility violation penalty
	float            _lambdaVs; // default = 100000

	// threshold for spatial regularization/smoothness term
	float            _tauSm; // default = 2

	// threshold for disparity-mapped photo-consistency term
	float            _tauPh; // default = 30

	// geometric bandwidth
	float            _sigmaX; // default = 20

	// pixel color bandwidth
	float            _sigmaC; // default = 4

	// disparity/depth bandwidth (for trilateral filter)
	float            _sigmaD; // default = 2

	// maximum neighborhood radius in pixels (the minimum spanning tree 
	// computation will become very slow if this is too large, 
	// i.e., > 40 pixels)
	int              _maxNbrRad; // default = 40 (pixels)

	// the number of overlapping minimum spanning trees used to construct each
	// pixel's neighborhood
	int              _numMst; // default = 2

	// the maximum number of iterations
	int              _maxIter; // default = 2

	// vector of output discrete disparity/depth map filename strings
	vector<string>   _dptFiles; // not required if _dptSmoothFiles is given

	// vector of output continuous/smooth disparity/depth map filename strings
	vector<string>   _dptSmoothFiles; // not required

	// vector of neighborhood edge set filename strings
	vector<string>   _nbrEdgeFiles; // default = neighborhood of (1,0), (0,1)

	// the inverse of the maximum depth
	double           _zMaxInv; // default = 0 (disp is zero, depth is inf)

	// the number of disparity/depth layers
	int              _numLayers; // default = 32

	// the scale each depth label in units of output depth map pixel intensity
	int              _depthScale; // default = 8

	// the initial disparity/depth label
	int              _initDptValue; // default = 0

	// noise removal threshold (ignored if _runNoiseremoval != !)
	int              _nrThresh; // default = 3

	////////////////// THE FOLLOWING ARE NOT SET BY THE USER //////////////////

	// vector of allowable labels
	vector<short>    _labels;

	// array of values that, when indexed, convert a disparity/depth label to
	// a depth value. this is not specified by the user.
	valarray<double> _lbl2invDepth;

	void Parse(
		int argc, 
		const char *argv[]);
};

#endif


