///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <valarray>
#include "energy.h"
#include "CmdArgParser.h"
#include "PVecMat.h"
#include "TVecMat.h"
#include "Image.h"
#include "Camera.h"
#include "mvstereo.h"
#include "expansion.h"

using std::valarray;


// Compute all energy terms.
double computeEnergyAll(
	const Parameters &params,
	const DataIn     &dataIn,
	DataOut          &dataOut)
{
	const int numCams  = int(dataIn._camArray.size());
	const int numPairs = int(params._camPairs.size());

	double esm = 0.0;
	double eph = 0.0;
	double evs = 0.0;

	for (int i = 0; i < numCams; ++i) 
	{
		if (params._lambda)
		{
			esm += computeEnergySm(
				params, 
				dataOut._lblArray[i], 
				dataIn._imArray[i], 
				dataIn._nbrEdgeArray[i], 
				dataIn._degFuncArray[i]);
		}
	}

	for (int i = 0; i < numPairs; ++i)
	{
		const int p = params._camPairs[i][0];
		const int q = params._camPairs[i][1];

		ImageOf<int> warpP2Q;
		ImageOf<int> warpQ2P;
		ImageOf<float> warpP2Qf;
	
		forwardWarpFieldInt(
			dataIn._camArray[p], 
			dataOut._lblArray[p], 
			INVALID_LABEL,
			params._lbl2invDepth, 
			dataIn._camArray[q], 
			warpP2Q);

		forwardWarpFieldInt(
			dataIn._camArray[q], 
			dataOut._lblArray[q], 
			INVALID_LABEL,
			params._lbl2invDepth, 
			dataIn._camArray[p], 
			warpQ2P);

		if(params._tauPh)
		{
			forwardWarpFieldFloat(
				dataIn._camArray[p], 
				dataOut._lblArray[p], 
				INVALID_LABEL,
				params._lbl2invDepth, 
				dataIn._camArray[q], 
				warpP2Qf);

			eph += computeEnergyPh(
				params, 
				dataIn._imArray[p], 
				dataIn._imArray[q], 
				dataIn._imArrayBound[p], 
				dataIn._imArrayBound[q],
				dataOut._lblArray[p], 
				dataOut._lblArray[q], 
				warpP2Q, 
				warpP2Qf);
		}			

		if(params._lambdaVs)
		{
			evs += computeEnergyVs(
				params, 
				dataOut._lblArray[p], 
				dataOut._lblArray[q], 
				warpP2Q);
			evs += computeEnergyVs(
				params, 
				dataOut._lblArray[q], 
				dataOut._lblArray[p], 
				warpQ2P);
		}

	}

	return eph + evs + esm;
}


// Compute the disparity-mapped photo-consistency energy term.
double computeEnergyPh(
	const Parameters     &params,
	const ImageOf<short> &imgP, 
	const ImageOf<short> &imgQ, 
	const ImageOf<short> &boundP, 
	const ImageOf<short> &boundQ, 
	const ImageOf<short> &lblP, 
	const ImageOf<short> &lblQ,
	const ImageOf<int>   &warpP2Q,
	const ImageOf<float> &warpP2Qf)
{
	double E = 0;	
	const Coord imgSizeQ(imgQ.Width(), imgQ.Height());
	const Coordf imgSizeQf(float(imgQ.Width()-1), float(imgQ.Height()-1));

	bool doBirchfield = false;
	if(params._birchfield == 1)
		doBirchfield = true;

	for (int y = 0; y < imgP.Height(); ++y)
	{
		for (int x = 0; x < imgP.Width(); ++x)
		{
			Coord p(x, y);
			float lp = lblP.Pixel(p.y, p.x);
			const int   *ptr_q  = warpP2Q.PixelAddress(p.y, p.x);
			const float *ptr_qf = warpP2Qf.PixelAddress(p.y, p.x);

			Coord  q(ptr_q[0] , ptr_q[1]);
			Coordf qf(ptr_qf[0], ptr_qf[1]);

			if (q  >= Coord(0,0)        && q  < imgSizeQ &&
				qf >= Coordf(0.0f,0.0f) && qf < imgSizeQf)
			{
				float lq = lblQ.Pixel(q.y, q.x);
				if (lp == lq)
				{
					if(doBirchfield)
					{
						E += penaltyPh(p, qf, imgP, boundP, 
						               imgQ, boundQ, params);
					}
					else
					{
						E += penaltyPh(p, qf, imgP, 
						               imgQ, params);
					}
				}
			}
		}
	}
	return E;
}


// Compute the regularization/smoothness energy term.
double computeEnergySm(
	const Parameters               &params,
	const ImageOf<short>           &lbl,
	const ImageOf<short>           &img,
	const ImageOf< vector<Coord> > &nbrEdgeSet,
	const ImageOf<float>           &dfArray)
{
	int height = img.Height();
	int width  = img.Width();

	const Coord imSize(width, height);

	double E = 0;

	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x)
		{
			Coord p(x,y);
			short lp = lbl.Pixel(p.y,p.x);
			vector<Coord> nbr = nbrEdgeSet.Pixel(p.y, p.x);
			
			for (size_t k=0; k < nbr.size(); k++)
			{
				Coord q = nbr[k];
				if ( q>=Coord(0,0) && q<imSize )
				{
					short lq = lbl.Pixel(q.y,q.x);
					E += penaltySm(p, q, lp, lq, img, dfArray, params);
				}
			}
		}
	}

	return E;
}


// Compute the visibility energy term.
double computeEnergyVs(
	const Parameters     &params,
	const ImageOf<short> &lblP,
	const ImageOf<short> &lblQ,
	const ImageOf<int>   &warpP2Q)
{
	double E = 0;
	Coord imgSizeQ(lblQ.Width(), lblQ.Height());

	for(int y = 0; y < lblP.Height(); y++)
	{
		for(int x = 0; x < lblP.Width(); x++)
		{
			Coord p(x, y);
			short lp = lblP.Pixel(p.y, p.x);
			const int *ptr_q = warpP2Q.PixelAddress(p.y, p.x);
			Coord q(ptr_q[0], ptr_q[1]);
			if(q >= Coord(0,0) && q < imgSizeQ)
			{
				short lq = lblQ.Pixel(q.y, q.x);
				E += penaltyVs(lp, lq, params);
			}
		}
	}

	return E;
}


/**
 * Create a graph corresponding to all of the non-alpha pixels in lbl, 
 * (the image containing all of the disparity labels) where the nodes are 
 * defined by the Energy class. If a pixel label is already alpha, then it 
 * is omitted from the graph (since it's alpha no matter what, regardless
 * of the result of a graph cut). A node is effectively omitted by 
 * setting it to "VAR_ACTIVE," which is the equivalent of NULL. 
 */
int addVariables(
	const ImageOf<short> &lbl,
	const short          &alpha,
	Energy               *energy,
	ImageOf<Energy::Var> &var) 
{
	int height = lbl.Height();
	int width  = lbl.Width();

	var.ReAllocate(height, width, 1);

	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x)
		{
			short l = lbl.Pixel(y,x);

			if (l == alpha)
			{
				// set the graph node to null (omit this node)
				var.Pixel(y,x) = VAR_ACTIVE;
			}
			else
			{
				// this is basically a pointer to a graph node
				var.Pixel(y,x) = energy->add_variable();
			}
		}
	}

	return 0;
}


// Add disparity-mapped photo-consistency terms to the graph.
// Data terms are interwined with visibility.
// warpP2Q is the size of source image, [0,1] is current depth, [2,3] is alpha;
int addTermsPh( 
	const Parameters           &params,
	const ImageOf<short>       &imgP, 
	const ImageOf<short>       &imgQ, 
	const ImageOf<short>       &boundP, 
	const ImageOf<short>       &boundQ, 
	const ImageOf<short>       &lblP, 
	const ImageOf<short>       &lblQ,
	const ImageOf<int>         &warpP2Q,
	const ImageOf<float>       &warpP2Qf,
	const short                &alpha,
	const ImageOf<Energy::Var> &varP,
	const ImageOf<Energy::Var> &varQ,
	Energy                     *energy)
{
	const Coord imgSizeQ(imgQ.Width(), imgQ.Height());
	const Coordf imgSizeQf(float(imgQ.Width()-1), float(imgQ.Height()-1));

	float E00 = 0;
	float E01 = 0;
	float E10 = 0;
	float E11 = 0;

	bool doBirchfield = false;
	if(params._birchfield == 1)
		doBirchfield = true;

	for (int y = 0; y < imgP.Height(); ++y)
	{
		for (int x = 0; x < imgP.Width(); ++x)
		{
			Coord p(x, y);
			short lp = lblP.Pixel(p.y, p.x);

			Energy::Var pvar = varP.Pixel(p.y, p.x);
			const int *ptr_q = warpP2Q.PixelAddress(p.y, p.x);
			const float *ptr_qf = warpP2Qf.PixelAddress(p.y, p.x);

			Coord q0(ptr_q[0], ptr_q[1]); // q mapped using original disparity
			Coord qA(ptr_q[2], ptr_q[3]); // q mapped using alpha disparity

			Coordf q0f(ptr_qf[0], ptr_qf[1]);
			Coordf qAf(ptr_qf[2], ptr_qf[3]);

			if (q0  >= Coord(0,0)         && q0  < imgSizeQ &&
				q0f >= Coordf(0.0f, 0.0f) && q0f < imgSizeQf)
			{
				short lq = lblQ.Pixel(q0.y, q0.x);
				Energy::Var qvar = varQ.Pixel(q0.y, q0.x);

				// CASE 1
				if(lp == lq)
				{
					if(doBirchfield)
					{
						E00 = penaltyPh(p, q0f, imgP, boundP, 
						                imgQ, boundQ, params);
					}
					else
					{
						E00 = penaltyPh(p, q0f, imgP, 
						                imgQ, params);
					}
					E01 = 0;
					E10 = 0;
					if(q0 == qA) E11 = E00; 
					else         E11 = 0;
					//E11 = (q0 == qA)*E00;

					if(lp == /* lq == */ alpha)
					{
						// no matter whether p and q are active, 
						// Epq = V(I1(p), I2(q))
						energy->add_constant(E11);
					}
					else // lp == lq != alpha
					{
						// both p and q are active
						energy->add_term2(pvar, qvar, E00, E01, E10, E11);
					}
				}

				// CASE 2
				else // lp != lq
				{
					E00 = 0;
					E01 = 0;
					E10 = 0;
					if(q0 == qA)
					{
						if(doBirchfield)
						{
							E11 = penaltyPh(p, q0f, imgP, boundP, 
							                imgQ, boundQ, params);
						}
						else
						{
							E11 = penaltyPh(p, q0f, imgP, 
							                imgQ, params);
						}
					}
					else
					{
						E11 = 0;
					}

					if(lp == alpha)
					{
						if(lq != alpha)
						{
							// q is active.
							// No matter whether p is active, Epq depends
							// only on xq
							energy->add_term1(qvar, E10, E11);
						}
					}
					else // lp != alpha
					{
						if(lq == alpha)
						{
							// p is active.
							// No matter whether q is active, Epq depends
							// only on xp
							energy->add_term1(pvar, E01, E11);
						}
						else // lq != alpha
						{
							// both p and q are active.
							energy->add_term2(pvar, qvar, E00, E01, E10, E11);
						}
					}
				}
			}

			// Note the test for q0 != qA. 
			if (q0  != qA && qA  >= Coord(0,0) && qA < imgSizeQ &&
				qAf >= Coordf(0.0f, 0.0f) && qAf <  imgSizeQf)
			{
				short lq = lblQ.Pixel(qA.y, qA.x);
				Energy::Var qvar = varQ.Pixel(qA.y, qA.x);

				// CASE 3
				if(lp == lq)
				{
					// p + alpha = q
					//if(lp /* == lq */ != alpha) <-- always true if q0 != qA
					//{
						// both p and q are active
						E00 = 0;
						E01 = 0;
						E10 = 0;
						if(doBirchfield)
						{
							E11 = penaltyPh(p, qAf, imgP, boundP, 
							                imgQ, boundQ, params);
						}
						else
						{
							E11 = penaltyPh(p, qAf, imgP, 
							                imgQ, params);
						}

						energy->add_term2(pvar, qvar, E00, E01, E10, E11);
					//}
					// else Epq = 0;
				}

				// CASE 4
				else // lp != lq
				{
					E00 = 0;
					E01 = 0;
					if(doBirchfield)
					{
						E11 = penaltyPh(p, qAf, imgP, boundP, 
						                imgQ, boundQ, params);
					}
					else
					{
						E11 = penaltyPh(p, qAf, imgP, 
						                imgQ, params);
					}
					if(alpha == lq) E10 = E11;
					else E10 = 0;
					//E10 = (alpha == lq)*E11;

					//if( lp != alpha )  <-- always true if q0 != qA
					//{
						// if( lq == alpha && lq.flow_u == alpha.flow_u && 
						//     lq.flow_v == alpha.flow_v )
						if( qvar == VAR_ACTIVE )
						{
							// q is inactive
							// Epq depends only on xp
							energy->add_term1(pvar, E01, E11);
						}
						// if( lq != alpha || lq.flow_u != alpha.flow_u || 
						//     lq.flow_v != alpha.flow_v )
						else
						{
							// q is active
							energy->add_term2(pvar, qvar, E00, E01, E10, E11);
						}
					//}
					// else Epq = 0;
				}
			}
		} // end for(int x ...
	} // end for(int y ...

	return 0;
}


// Add spatial regularization/smoothness terms to the graph.
int addTermsSm(
	const Parameters               &params,
	const ImageOf<short>           &img,
	const ImageOf<short>           &lbl,
	const short                    &alpha,
	const ImageOf<Energy::Var>     &vid,
	const ImageOf< vector<Coord> > &nbrEdgeSet,
	const ImageOf<float>           &dfArray,
	Energy                         *energy)
{
	int height = img.Height();
	int width  = img.Width();

	const Coord imSize(width, height);

	float E00 = 0;
	float E01 = 0;
	float E10 = 0;
	float E11 = 0;

		for (int y = 0; y < height; ++y)
		{
			for (int x = 0; x < width; ++x)
			{
				Coord p(x,y);
				short lp = lbl.Pixel(p.y, p.x);
				Energy::Var pvar = vid.Pixel(p.y, p.x);

				vector<Coord> nbr = nbrEdgeSet.Pixel(p.y, p.x);

				for (unsigned int k=0; k < nbr.size(); k++)
				{
					Coord q = nbr[k];

					short lq = lbl.Pixel(q.y, q.x);
					Energy::Var qvar = vid.Pixel(q.y, q.x);

					E00 = penaltySm(p, q, lp, lq, img, dfArray, params);
					E01 = penaltySm(p, q, lp, alpha, img, dfArray, params);
					E10 = penaltySm(p, q, alpha, lq, img, dfArray, params);
					E11 = penaltySm(p, q, alpha, alpha, img, dfArray, params);

					if (pvar != VAR_ACTIVE)
					{
						if (qvar != VAR_ACTIVE)
						{
							energy -> add_term2(pvar, qvar, 
								E00, E01, E10, E11);
						}
						else
						{
							energy -> add_term1(pvar, E01, E11);
						}
					}
					else
					{
						if (qvar != VAR_ACTIVE)
						{
							energy -> add_term1(qvar, E10, E11);
						}
						else
						{
							energy -> add_constant(E11);
						}
					}
				}
			}
		}

	return 0;
}


// Add visibility terms to the graph.
int addTermsVs(
	const Parameters           &params,
	const ImageOf<short>       &lblP,
	const ImageOf<short>       &lblQ, 
	const ImageOf<int>         &warpP2Q,
	const short                &alpha,
	const ImageOf<Energy::Var> &varP,
	const ImageOf<Energy::Var> &varQ,
	Energy                     *energy)
{
	Coord imgSizeQ(lblQ.Width(), lblQ.Height());
	
	float E00 = 0;
	float E01 = 0;
	float E10 = 0;
	float E11 = 0;

	for(int y = 0; y < lblP.Height(); y++)
	{
		for(int x = 0; x < lblP.Width(); x++)
		{
			Coord p(x, y);
			short lp = lblP.Pixel(p.y, p.x);

			Energy::Var pvar = varP.Pixel(p.y, p.x);
			const int *ptr_q = warpP2Q.PixelAddress(p.y, p.x);

			Coord q0(ptr_q[0], ptr_q[1]);
			Coord qA(ptr_q[2], ptr_q[3]);
			if(q0 >= Coord(0,0) && q0 < imgSizeQ)
			{
				short lq = lblQ.Pixel(q0.y, q0.x);
				Energy::Var qvar = varQ.Pixel(q0.y, q0.x);

				E00 = penaltyVs(lp, lq, params);
				E01 = penaltyVs(lp, alpha, params);
				if(q0 == qA)
				{
					E10 = penaltyVs(alpha, lq, params);
					E11 = penaltyVs(alpha, alpha, params);
				}
				else
				{
					E10 = 0;
					E11 = 0;
				}
				//E10 = (q0 == qA)*penaltyVs(alpha, lq, params);
				//E11 = (q0 == qA)*penaltyVs(alpha, alpha, params);

				if(lp == alpha)
				{
					if(lq == alpha)
					{
						// no matter if p and q are active or not
						energy->add_constant(E11);
					}
					else // lq != alpha
					{
						// q is active.
						// No matter if p is active or not,
						// Epq depends only on xq
						energy->add_term1(qvar, E10, E11);
					}
				}
				else // lp != alpha 
				{
					if(lq == alpha)
					{
						// p is active.
						// No matter if q is active or not,
						// Epq depends only on xp
						energy->add_term1(pvar, E01, E11);
					}
					else // lq != alpha
					{
						// both p and q are active
						energy->add_term2(pvar, qvar, E00, E01, E10, E11);
					}
				}
			}

			if(q0 != qA && qA >= Coord(0,0) && qA < imgSizeQ)
			{
				short lq = lblQ.Pixel(qA.y, qA.x);
				Energy::Var qvar = varQ.Pixel(qA.y, qA.x);

				// because q0 != qA, lp != alpha, which
				// means p is active

				E00 = 0;
				E01 = 0;
				E10 = penaltyVs(alpha, lq, params);
				E11 = penaltyVs(alpha, alpha, params);

				//if(lq == alpha && lq.flow_u == alpha.flow_u && 
				//   lq.flow_v == alpha_flow_v)
				if(qvar == VAR_ACTIVE)
				{
					// q is inactive because lq == alpha
					// Epq only depends on xp
					energy->add_term1(pvar, E01, E11);
				}
				// else  lq != alpha || lq.flow_u != alpha.flow_u || 
				//       lq.flow_v != alpha_flow_v)
				else 
				{
					// q is active
					energy->add_term2(pvar, qvar, E00, E01, E10, E11);
				}
			}
		}
	}

	return 0;
}


double expandAlpha(
	const Parameters &params,
	const DataIn     &dataIn,
	const short      alpha,
	const double     E_old,
	DataOut          &dataOut)
{
	//assume out->_lblArray is initialized properly

	Energy *energy = new Energy();

	const int numCams   = int(dataIn._camArray.size());
	const int numPairs  = int(params._camPairs.size());
	const int height    = dataOut._lblArray[0].Height();
	const int width     = dataOut._lblArray[0].Width();

	// the smoothness terms are computed from each image separately
	for (int i = 0; i < numCams; ++i) 
	{
		addVariables(
			dataOut._lblArray[i], 
			alpha, 
			energy, 
			dataOut._varArray[i]);

		if (params._lambda)
		{
			addTermsSm(
				params, 
				dataIn._imArray[i], 
				dataOut._lblArray[i],
				alpha, 
				dataOut._varArray[i], 
				dataIn._nbrEdgeArray[i], 
				dataIn._degFuncArray[i], 
				energy);
		}
	}

	// the data terms are computed from *pairs* of images
	for (int i = 0; i < numPairs; ++i)
	{
		const int p = params._camPairs[i][0];
		const int q = params._camPairs[i][1];

		ImageOf<int>   warpP2Q;
		ImageOf<int>   warpQ2P;
		ImageOf<float> warpP2Qf;

		forwardWarpFieldInt(
			dataIn._camArray[p], 
			dataOut._lblArray[p], 
			alpha,
			params._lbl2invDepth, 
			dataIn._camArray[q], 
			warpP2Q);

		forwardWarpFieldInt(
			dataIn._camArray[q], 
			dataOut._lblArray[q],
			alpha,
			params._lbl2invDepth, 
			dataIn._camArray[p], 
			warpQ2P);

		if(params._tauPh)
		{
			forwardWarpFieldFloat(
				dataIn._camArray[p], 
				dataOut._lblArray[p], 
				alpha,
				params._lbl2invDepth, 
				dataIn._camArray[q], 
				warpP2Qf);

			addTermsPh(
				params, 
				dataIn._imArray[p], 
				dataIn._imArray[q], 
				dataIn._imArrayBound[p], 
				dataIn._imArrayBound[q],
				dataOut._lblArray[p], 
				dataOut._lblArray[q],
				warpP2Q, 
				warpP2Qf,
				alpha,
				dataOut._varArray[p], 
				dataOut._varArray[q], 
				energy); 
		}

		if(params._lambdaVs)
		{
			addTermsVs(
				params, 
				dataOut._lblArray[p], 
				dataOut._lblArray[q], 
				warpP2Q, 
				alpha, 
				dataOut._varArray[p], 
				dataOut._varArray[q], 
				energy);

			addTermsVs(
				params, 
				dataOut._lblArray[q], 
				dataOut._lblArray[p], 
				warpQ2P, 
				alpha, 
				dataOut._varArray[q], 
				dataOut._varArray[p], 
				energy);
		}
	}

	//compute energy
	double E = energy -> minimize();

	if (E < E_old)
	{
		for (int i = 0; i < numCams; ++i)
		{
			for (int y = 0; y < height; ++y)
			{
				for (int x = 0; x < width; ++x)
				{
					Energy::Var var2 = dataOut._varArray[i].Pixel(y,x);

					// if the current node isn't NULL (i.e., it's included
					// in the graph) and the current alpha expansion resulted
					// in the t-link from node to alpha being cut, then set
					// the corresponding label to alpha. 
					if (var2 != VAR_ACTIVE && energy->get_var(var2) == VALUE1)
					{
						dataOut._lblArray[i].Pixel(y,x) = alpha;
					}
				}
			}
		}
	}

	delete energy;
	return E;
}
