///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <time.h>
#include <valarray>
#include <fstream>
#include <omp.h>
using std::valarray;

#include "data.h"
#include "mvstereo.h"
#include "mstnbr.h"
#include "trilatfilt.h"
#include "PVecMat.h"


int main(int argc, const char *argv[])
{
	////////////////////// READ AND PARSE ARGUMENTS FILE //////////////////////

	// holds all of the input parameters and filenames of the input data
	Parameters params;
	// reads in the input parameters from the arguments file (af)
	params.Parse(argc, argv);


	///////////////////////////// LOAD INPUT DATA /////////////////////////////

	printf("Allocating space and loading images...\n");

	DataIn dataIn;
	DataOut dataOut;

	const int numCams = (int)params._camFiles.size();
	const int numImgs = (int)params._imFiles.size();
	const int numPairs = (int)params._camPairs.size();

	if(numCams != numImgs)
	{
		printf("Number of cameras does not match number of images!\n");
		exit(0);
	}

	dataIn._camArray.resize(numCams);
	dataIn._imArray.resize(numCams);
	dataIn._imArrayBound.resize(numCams);
	dataIn._degFuncArray.resize(numCams);
	dataIn._nbrEdgeArray.resize(numCams);

	dataOut._lblArray.resize(numCams);
	dataOut._dptSmoothArray.resize(numCams);
	dataOut._varArray.resize(numCams);

	//load images
	for (int i = 0; i < numImgs; i++)
	{		
		string fname = params._camFiles[i];
		dataIn._camArray[i].LoadFromFile(fname.c_str());
		dataIn._camArray[i].Parameter2Matrix();

		fname = params._imFiles[i];
		if (!LoadPixels(fname.c_str(), dataIn._imArray[i])) exit(0);

		int height = dataIn._imArray[i].Height();
		int width  = dataIn._imArray[i].Width();

		//compute image bounds for fast cost evaluation
		if (params._birchfield)
		{
			computeImageBound(dataIn._imArray[i], dataIn._imArrayBound[i]);
		}
		else
		{
			CopyPixels(dataIn._imArrayBound[i], dataIn._imArray[i]);
		}

		dataOut._lblArray[i].ReAllocate(height, width, 1);
		for(int y = 0; y < height; y++)
		{
			for(int x = 0; x < width; x++)
			{
				dataOut._lblArray[i].Pixel(y,x) = params._initDptValue;
			}
		}	
	} // end for( int i...


	////////////// MINIMUM SPANNING TREE NEIGHBORHOOD COMPUTATION /////////////

	// if the arguments file indicates that the minimum spanning tree (mst)
	// neighborhood (nbr) computation should be performed...
	if(params._runMstnbr == 1)
	{
		// run multiple threads since each image can be processed 
		// independently
		#pragma omp parallel for num_threads(params._numThreads)
			for(int i = 0; i < numImgs; i++)
			{
				computeNbrMST(params, i, dataIn);
				writeNbrMST(params._nbrEdgeFiles[i], dataIn._nbrEdgeArray[i]);
			}
	}
	// else read in precomputed minimum spanning tree neighborhoods)...
	else
	{
		// only need to read in MST neighborhood if performing stereo matching
		if(params._runMvstereo == 1)
		{
			printf("Reading MST neighborhood file...\n");

			for (int i = 0; i < numImgs; i++)
			{
				int height = dataIn._imArray[i].Height();
				int width  = dataIn._imArray[i].Width();
				string fname = params._nbrEdgeFiles[i];

				dataIn._nbrEdgeArray[i].ReAllocate(height, width, 1);
				// if fname is empty, then the neighborhood will default to 
				// (0,1), (1,0)
				readNbrMST(fname, dataIn._nbrEdgeArray[i]);
			}
		}
	}


	////////////////////// MULTIPLE VIEW STEREO ALGORITHM /////////////////////

	// if the arguments file indicates that the multiple-view stereo
	// algorithm should be run... 
	if(params._runMvstereo == 1)
	{
		// compute the degree function
		#pragma omp parallel for num_threads(params._numThreads)
			for (int i = 0; i < numImgs; i++)
			{
				computeDegFuncArray(params, i, dataIn); 
			}

		multiViewMatch(params, dataIn, dataOut);
	}


	///////////////// DISPARITY/DEPTH MAP NOISE REMOVAL PROCESS ///////////////
	if(params._runNoiseremoval == 1)
	{
		#pragma omp parallel for num_threads(params._numThreads)
			for(int i = 0; i < numImgs; i++)
			{
				if(params._runMvstereo == 0)
				{
					// load depth maps from the file system
					ImageOf<double> tmpDepth;
					bool success 
						= LoadPixels(params._dptFiles[i].c_str(), tmpDepth);

					if (!success) 
					{
						printf("WARNING: Disparity/depth map %d not computed (runMvstereo == 0)!\n", i);
						printf("   The disparity/depth file does not exist either.\n");
						printf("   Skipping noise removal process despite option -runNoiseremoval == 1.\n");
						continue;
					}
					else
					{
						int height = tmpDepth.Height();
						int width  = tmpDepth.Width();
						double scale = params._depthScale;
						for(int y = 0; y < height; y++)
						{
							for(int x = 0; x < width; x++)
							{
								dataOut._lblArray[i].Pixel(y,x) = (short)round(
									tmpDepth.Pixel(y,x) / scale);
							}
						}
					}
				}

				removeNoise(params, i, dataOut);
			}
	}

	// save results
	// note: this will overwrite the noisy depth maps if they
	// were previously saved to disk
	if(params._runMvstereo == 1 || params._runNoiseremoval == 1)
	{
		for (int i = 0; i < numImgs; i++)
		{
			string fname = params._dptFiles[i];

			if(fname.empty())
			{
				if(params._runTrilatfilt == 0 || 
				   params._dptSmoothFiles[i].empty())
				{
					printf("WARNING: No discrete or continuous disparity/depth map filenames found for image %d!\n");
					printf("   The result will not be saved.\n", i);
				}
				continue;
			}

			int height = dataOut._lblArray[i].Height();
			int width  = dataOut._lblArray[i].Width();
			ImageOf<short> tmpDepth;
			tmpDepth.ReAllocate(height, width, 1);
			for(int y = 0; y < height; y++)
			{
				for(int x = 0; x < width; x++)
				{
					tmpDepth.Pixel(y,x) = dataOut._lblArray[i].Pixel(y,x);
				}
			}

			multiply_assign(tmpDepth.FirstPixel(), 
				(short)params._depthScale, 
				tmpDepth.NumImgPixels());
			SavePixels(fname.c_str(), tmpDepth);
		} // end for( int i...
	}


	////////////////////// TRILATERAL FILTER POST PROCESS /////////////////////

	// if the arguments file indicates that trilateral filtering should be 
	// performed as a post-processing step... 
	if(params._runTrilatfilt == 1)
	{
		// run multiple threads since each image can be processed independently
		#pragma omp parallel for num_threads(params._numThreads)
			for(int i = 0; i < numImgs; i++)
			{
				string fname = params._dptSmoothFiles[i];
				if(fname.empty())
				{
					printf("WARNING: Continuous/smooth disparity/depth map filename %d not found!\n", i);
					printf("   Skipping trilateral filtering despite option -run_trilatfilter = 1.\n");
					continue;
				}

				if(params._runMvstereo == 0 && params._runNoiseremoval == 0)
				{
					// the depth maps have not yet been opened
					ImageOf<double> tmpDepth;
					bool success 
						= LoadPixels(params._dptFiles[i].c_str(), tmpDepth);

					if (!success) 
					{
						printf("WARNING: Disparity/depth map %d not computed (runMvstereo == 0)!\n", i);
						printf("   The disparity/depth file does not exist either.\n");
						printf("   Skipping trilateral filter despite option -runTrilatfilt == 1.\n");
						continue;
					}
					else
					{
						int height = tmpDepth.Height();
						int width  = tmpDepth.Width();
						double scale = params._depthScale;
						for(int y = 0; y < height; y++)
						{
							for(int x = 0; x < width; x++)
							{
								dataOut._lblArray[i].Pixel(y,x) = (short)round(
									tmpDepth.Pixel(y,x) / scale);
							}
						}
					}
				}

				trilateralFilter(params, i, dataIn,	dataOut);

				// save results
				double scale = params._depthScale;
				int height = dataOut._dptSmoothArray[i].Height();
				int width  = dataOut._dptSmoothArray[i].Width();
				ImageOf<short> tmpDepth;
				tmpDepth.ReAllocate(height, width, 1);
				for(int y = 0; y < height; y++)
				{
					for(int x = 0; x < width; x++)
					{
						tmpDepth.Pixel(y,x) = (short)round(
							dataOut._dptSmoothArray[i].Pixel(y,x) * scale);
					}
				}
				SavePixels(fname.c_str(), tmpDepth);
			}
	}

	printf("Done.\n");

	return 0;
}

