///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "mstnbr.h"


// Main function for computing the minimum spanning tree neighborhood structure 
void computeNbrMST(
	const Parameters &params, 
	const int        imageIndex,
	DataIn           &dataIn)
{
	printf("Computing MST neighborhood for image %d...\n", imageIndex+1);

	const double  sgmg = params._sigmaX;     // geometric sigma/bandwidth
	const double  sgmc = params._sigmaC;     // pixel color sigma/bandwidth
	const int   radius = params._maxNbrRad; // maximum neighborhood radius
	const int numTrees = params._numMst;     // number of minimum spanning trees
	const int   height = dataIn._imArray[imageIndex].Height();
	const int    width = dataIn._imArray[imageIndex].Width();

	dataIn._nbrEdgeArray[imageIndex].ReAllocate(height,width,1);

	valarray<Vec3int> nbr((2*radius+1)*(2*radius+1)-1);
	int count = 0;
	for (int y = -radius; y<=radius; ++y)
	{
		for (int x = -radius; x<=radius; ++x)
		{
			if (x==0&&y==0) continue;

			nbr[count][0] = x;
			nbr[count][1] = y;
			nbr[count][2] = x*x+y*y;
			count++;
		}
	}

	// at each pixel p, save a set of q where p<q (in terms of index order), 
	// such that p--q is a selected edge in one of the trees

	ImageOf< vector<int> > edgeMap;

	mstPrim(params, dataIn._imArray[imageIndex], nbr, edgeMap);

	//count how many edges in total and write them to dataIn._nbrEdgeArray[imageIndex];
	count = 0;
	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			const vector<int> &vec = edgeMap.Pixel(y,x);
			int vecSize = int(vec.size());
			count += vecSize;

			// for each q in hs,
			// let yy = q/imgW and xx = q%imgW,
			// there should be an edge p-q with weight 
			// wpq = k(p,q)*(1/sumk(p)+1/sumk(q));

			for(int i = 0; i < vecSize; i++)
			{
				int nx = vec[i] % width;
				int ny = vec[i] / width;
				dataIn._nbrEdgeArray[imageIndex].Pixel(y,x).push_back( Coord(nx, ny) );
			}
		}
	}
}


// Prim's minimum spanning tree algorithm
void mstPrim(
	const Parameters         &params,
	const ImageOf<short>     &img,
	const valarray<Vec3int>  &nbr,
	ImageOf< vector<int> >   &edgeMap)
{
	// MST-PRIM algorithm, pp509, Intro to Algorithms.

	const double sgmg   = params._sigmaX;
	const double sgmc   = params._sigmaC;
	const int numTrees  = params._numMst;
	const int imgH      = img.Height();
	const int imgW      = img.Width();
	const int numPixels = imgH*imgW;
	const int numNBRs = (int)nbr.size();

	const double fx = 1/(2*sgmg*sgmg);
	const double fc = 1/(2*sgmc*sgmc);

	// at pixel (x,y), it saves the min exponent within the neighborhood
	ImageOf<double> minD; 

	// at pixel (x,y), it saves the sum of exp(-(d-minD))
	ImageOf<double> sumK; 

	minD.ReAllocate(imgH, imgW, 1);
	sumK.ReAllocate(imgH, imgW, 1);

	minD.Fill(DBL_MAX);
	sumK.Fill(0);

	// set up minD and sumK;
	for (int y=0; y < imgH; ++y)
	{
		for (int x=0; x < imgW; ++x)
		{
			double minD_this = DBL_MAX;
			double sumK_this = 0;

			const short *p = img.PixelAddress(y,x);
			for (int n=0; n < numNBRs; ++n)
			{
				int xx = x + nbr[n][0];
				int yy = y + nbr[n][1];
				if (0<=xx && xx<imgW && 0<=yy && yy<imgH)
				{
					const short *q = img.PixelAddress(yy,xx);
					const double deltaC = colorDist(p, q);
					const double deltaX = nbr[n][2];
					const double D = fc*deltaC+fx*deltaX;
					if (D<minD_this)
					{
						const double fac = exp(-(minD_this-D));
						const double kpq = 1;
						sumK_this = fac*sumK_this+kpq;
						minD_this = D;
					}
					else
					{
						const double kpq = exp(-(D-minD_this));
						sumK_this += kpq;
					}
				}
			}
			minD.Pixel(y,x) = minD_this;
			sumK.Pixel(y,x) = sumK_this;
		}
	}

	// now search for the tree;
	edgeMap.ReAllocate(imgH, imgW, 1);

	// this is pi in the pseudo code, each pixel saves its parent pixel index 
	// (y*imgW+x)
	ImageOf<int> mapPar; 

	// a mask that remembers which node has bee selected in the tree
	ImageOf<unsigned char> selected; 

	mapPar.ReAllocate(imgH, imgW, 1);
	selected.ReAllocate(imgH, imgW, 1);

	PixelKey **pkHeap = new PixelKey*[numPixels];

	const double MAX_WEIGHT = 10;
	PixelKey *pkBuf = new PixelKey[numPixels];

	for (int t = 0; t < numTrees; ++t)
	{
		mapPar.Fill(-1);
		selected.Fill(0);

		for (int y = 0; y < imgH; ++y)
		{
			for (int x = 0; x < imgW; ++x)
			{
				int i = y*imgW+x;
				pkBuf[i]._key = MAX_WEIGHT;
				pkBuf[i]._index = i;
				pkHeap[i] = pkBuf+i;
			}
		}

		int pqSize = numPixels; //initial size of the priority queue
		int numComponents = 0;
		double totalWeight = 0;

		int pct = 0;
		while (pqSize)
		{
			const PixelKey &thisPK = pkBuf[0];
			int y = (thisPK._index/imgW);
			int x = (thisPK._index%imgW);
			int i = thisPK._index;
			const short *p = img.PixelAddress(y,x);

			selected.Pixel(y,x) = 1;

			// not connedted to the existing tree node, a new root
			if (mapPar.Pixel(y,x)==-1) 
			{
				++numComponents;
				if (thisPK._key!=MAX_WEIGHT) printf("something is wrong\n");
			}
			else
				totalWeight += thisPK._key;

			// these two have to be put in front of the for loop;
			// otherwise the bubble up will change the top of the heap.
			Heap1<PixelKey>::PopMax(pkHeap, pkBuf, pqSize);
			pqSize--;

			for (int n=0; n < numNBRs; ++n)
			{
				int xx = x + nbr[n][0];
				int yy = y + nbr[n][1];

				if (0<=xx && xx<imgW && 0<=yy && yy<imgH)
				{
					// the pixel is not seleted into the tree yet.
					if (selected.Pixel(yy,xx)==0) 
					{
						int ii = yy*imgW+xx;
						//bool good = false;
						bool good = true;
						if (i<ii)
						{
							const vector<int> &vec = edgeMap.Pixel(y,x);
							int vecSize = int(vec.size());
							for(int v = 0; v < vecSize; v++)
							{
								if(vec[v] == ii)
								{
									good = false;
									break;
								}
							}
						}
						else if (ii<i)
						{
							const vector<int> &vec = edgeMap.Pixel(yy,xx);
							int vecSize = int(vec.size());
							for(int v = 0; v < vecSize; v++)
							{
								if(vec[v] == i)
								{
									good = false;
									break;
								}
							}
						}
						else
						{
							printf("invalid edge config!\n");
							exit(0);
						}

						if (good)
						{
							const short *q = img.PixelAddress(yy,xx);

							// compute the weight between p and q;
							const double deltaC = colorDist(p, q);
							const double deltaX = nbr[n][2];
							const double D = fc*deltaC+fx*deltaX;

							const double kpq 
								= exp(-(D-minD.Pixel(y,x)))/sumK.Pixel(y,x);
							const double kqp 
								= exp(-(D-minD.Pixel(yy,xx)))/sumK.Pixel(yy,xx);
							// both kpq and kqp are within [0,1],
							// we wish to use edges with large kpq+kqp, 
							// but we do so by searching for min spanning tree
							// so we use 2-(kpq+kqp) as weight
							const double w = 2-(kpq+kqp);

							int k = pkHeap[ii] - pkBuf;
							// update key and adjust the priority queue 
							// (implemented as a heap)
							if (w < pkBuf[k]._key) 
							{
								pkBuf[k]._key = w;
								Heap1<PixelKey>::BubbleUp(pkHeap, pkBuf, k);
								mapPar.Pixel(yy, xx) = i;
							}
						}
					}
				}
			}
		} // end while (pqSize)

		//add the tree to the edge map
		int count = 0;
		for (int y = 0; y < imgH; ++y)
		{
			for (int x = 0; x < imgW; ++x)
			{
				int i = y*imgW+x;
				int pi = mapPar.Pixel(y,x);
				if (pi!=-1) //not a root
				{
					count++;
					bool good = true;
					if (pi<i)
					{
						vector<int> &vec = edgeMap.Pixel(pi/imgW,pi%imgW);
						int vecSize = int(vec.size());
						for(int v = 0; v < vecSize; v++)
						{
							if(vec[v] == i)
							{
								good = false;
								break;
							}
						}
						if(good)
						{
							vec.push_back(i);
						}
					}
					else if (pi>i)
					{
						vector<int> &vec = edgeMap.Pixel(i/imgW, i%imgW);
						int vecSize = int(vec.size());
						for(int v = 0; v < vecSize; v++)
						{
							if(vec[v] == pi)
							{
								good = false;
								break;
							}
						}
						if(good)
						{
							vec.push_back(pi);
						}
					}
					else
					{
						printf("Invalid edge config!\n");
						exit(0);
					}
				}
			}
		}
		printf("...tree %d of %d is done, total weight %lf\n", t+1, numTrees, totalWeight);
	}

	delete[] pkBuf;
}


// Read MST neighborhood file
void readNbrMST(
	const string             &filename, 
	ImageOf< vector<Coord> > &nbrMST)
{
	// It is assumed that nbrMST has already been allocated
	int height = nbrMST.Height();
	int width  = nbrMST.Width();

	nbrMST.ReAllocate(height,width,1);

	if(filename.empty() || filename.compare("none") == 0)
	{
		printf("WARNING: No neighborhood edge set provided.\n");
		printf("   Using (0,1), (1,0) as heighborhood.\n");

		for(int y = 0; y < height-1; y++)
		{
			for(int x = 0; x < width-1; x++)
			{
				if(1+y < height)
					nbrMST.Pixel(y,x).push_back( Coord(x,1+y) );
				if(1+x < width)
					nbrMST.Pixel(y,x).push_back( Coord(1+x,y) );
			}
		}
	}
	else
	{
		ifstream fin(filename.c_str(), ifstream::in);

		if( !fin.good() )
		{
			printf("Cannot open %s!\n", filename.c_str());
			exit(0);
		}

		string line;

		while( getline(fin, line) )
		{
			string::size_type start = 0;
			string::size_type end   = 0;

			int px = 0;
			int py = 0;

			string::size_type str_length = line.length();

			while(start < str_length)
			{
				vector<Coord> v;

				end = line.find(' ', start);

				if(end == string::npos)
				{
					end = str_length;
				}

				int pindex = atoi(line.substr(start, end-start).c_str());
				int x = pindex % width;
				// naturally performs floor(pindex / width);
				int y = pindex / width; 

				if(start == 0)
				{
					px = x;
					py = y;
				}
				else
				{
					nbrMST.Pixel(py, px).push_back( Coord(x, y) );
				}

				start = end + 1;
			}
		}

		fin.close();
	}
}


void writeNbrMST(
	const string                   &filename, 
	const ImageOf< vector<Coord> > &nbrMST)
{
	// if the file name string is empty, then don't save the edge set
	if( filename.empty() )
	{
		return;
	}

	int height = nbrMST.Height();
	int width  = nbrMST.Width();

	FILE *fp = fopen(filename.c_str(), "w");

	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			fprintf(fp, "%d", y*width+x);
			vector<Coord> nbrs = nbrMST.Pixel(y,x,0);
			int numNbrs = (int)nbrs.size();
			for(int n = 0; n < numNbrs; n++)
			{
				int index = nbrs[n].y * width + nbrs[n].x;
				fprintf(fp, " %d", index);
			}
			fprintf(fp, "\n");
		}
	}

	fclose(fp);
}


void swap(
	PixelKey **heap, 
	PixelKey &left, 
	PixelKey &right)
{
	double tmpKey;
	tmpKey = left._key;
	left._key = right._key;
	right._key = tmpKey;

	int tmpIdx;
	tmpIdx = left._index;
	left._index = right._index;
	right._index = tmpIdx;

//	swap(left._key, right._key);
//	swap(left._index, right._index);

	heap[left._index] = &left;
	heap[right._index] = &right;
}


void copy(
	PixelKey **heap, 
	PixelKey &left, 
	const PixelKey &right)
{
	left._key = right._key;
	left._index = right._index;
	
	heap[left._index] = &left;
}
