///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <valarray>
#include <sstream>
#include "energy.h"
#include "CmdArgParser.h"
#include "PVecMat.h"
#include "TVecMat.h"
#include "Image.h"
#include "Camera.h"
#include "mvstereo.h"
#include "expansion.h"
#include "mstnbr.h"

using std::valarray;


void multiViewMatch(
	const Parameters &params,
	const DataIn &dataIn, 
	DataOut &dataOut)
{
	printf("Running multi-view stereo matching algorithm...\n");

	int numLabels = (int)params._labels.size();

	double E = computeEnergyAll(params, dataIn, dataOut);
	printf("Initial Energy =                        %lf\n", E);

	int numExpansions = 0;
	int counter = 0;
	for (int iter = 0; iter < params._maxIter; ++iter)
	{
		for (int index = 0; index < numLabels; index++)
		{
			int disp_lbl = params._labels[index];
			double E_old = E;

			E = expandAlpha(params, dataIn, disp_lbl, E_old, dataOut);
			
			printf("Iter %02d of %02d, Layer %02d of %02d, Energy = %lf\n", iter+1, params._maxIter, index+1, numLabels, E);
			fflush(stdout);

			++numExpansions;
		}
	}
}


void computeImageBound(
	const ImageOf <short> &img,
	ImageOf <short> &bd)
{
	if (img.NBands()!=3)
	{
		fprintf(stderr, "computeImageBound assumes RGB image input!\n"); 
		exit(0);
	}

	bd.ReAllocate(img.Height(), img.Width(), 6);
	for (int y = 0; y < img.Height(); ++y)
	{
		for (int x = 0; x < img.Width(); ++x)
		{
			const short *color = img.PixelAddress(y,x);
			short    *minColor = bd.PixelAddress(y,x,0);
			short    *maxColor = bd.PixelAddress(y,x,3);

			minColor[0] = color[0];
			minColor[1] = color[1];
			minColor[2] = color[2];			
			maxColor[0] = color[0];
			maxColor[1] = color[1];
			maxColor[2] = color[2];

			if (x > 0) 
				boundingColor(minColor, maxColor, color, img.PixelAddress(y,x-1));
			if (x < img.Width()-1) 
				boundingColor(minColor, maxColor, color, img.PixelAddress(y,x+1));
			if (y > 0)
				boundingColor(minColor, maxColor, color, img.PixelAddress(y-1,x));
			if (y < img.Height()-1) 
				boundingColor(minColor, maxColor, color, img.PixelAddress(y+1,x));
		}
	}
}


void forwardWarpFieldInt(
	const Camera &srcCam, 
	const ImageOf <short> &srcLabel,
	const short alpha_disp_lbl,
	const valarray<double> &lbl2invDepth, //lbl2Depth is in the reference view!!!
	const Camera &dstCam,
	ImageOf<int> &warpField)
{
	const int srcH = srcLabel.Height();
	const int srcW = srcLabel.Width();

	if (alpha_disp_lbl == INVALID_LABEL)
		warpField.ReAllocate(srcH, srcW, 2);
	else 
		warpField.ReAllocate(srcH, srcW, 4);

	SqMat3double M_src = inverse(srcCam._intrinsicMat*srcCam._rotation);
	Vec3double V_src = srcCam._translation*srcCam._rotation;

	SqMat3double M_dst = dstCam._intrinsicMat*dstCam._rotation;//*inverse(refCam._intrinsicMat);
	Vec3double V_dst = dstCam._intrinsicMat*dstCam._translation;

	for (int y = 0; y < srcH; ++y)
	{
		for (int x = 0; x < srcW; ++x)
		{
			Vec3double x_src, x_dst;
			x_src[0] = x;
			x_src[1] = y;
			x_src[2] = 1;

			twoStepWarp(x_src, lbl2invDepth[srcLabel.Pixel(y,x)], 
				M_src, V_src, M_dst, V_dst, x_dst);

			int *q = warpField.PixelAddress(y,x);
			
			q[0] = round2(x_dst[0]);
			q[1] = round2(x_dst[1]);

			if (alpha_disp_lbl != INVALID_LABEL)
			{
				twoStepWarp(x_src, lbl2invDepth[alpha_disp_lbl], 
					M_src, V_src, M_dst, V_dst, x_dst);
			
				q[2] = round2(x_dst[0]);
				q[3] = round2(x_dst[1]);
			}
		}
	}
}


// returns a floating point warp instead of int
void forwardWarpFieldFloat(
	const Camera &srcCam, 
	const ImageOf <short> &srcLabel,
	const short alpha_disp_lbl,
	const valarray<double> &lbl2invDepth, //lbl2Depth is in the reference view!!!
	const Camera &dstCam,
	ImageOf<float> &warpField)
{
	const int srcH = srcLabel.Height();
	const int srcW = srcLabel.Width();

	if (alpha_disp_lbl == INVALID_LABEL)
		warpField.ReAllocate(srcH, srcW, 2);
	else 
		warpField.ReAllocate(srcH, srcW, 4);

	SqMat3double M_src = inverse(srcCam._intrinsicMat*srcCam._rotation);
	Vec3double V_src = srcCam._translation*srcCam._rotation;

	SqMat3double M_dst = dstCam._intrinsicMat*dstCam._rotation;//*inverse(refCam._intrinsicMat);
	Vec3double V_dst = dstCam._intrinsicMat*dstCam._translation;

	for (int y = 0; y < srcH; ++y)
	{
		for (int x = 0; x < srcW; ++x)
		{
			Vec3double x_src, x_dst;
			x_src[0] = x;
			x_src[1] = y;
			x_src[2] = 1;

			twoStepWarp(x_src, lbl2invDepth[srcLabel.Pixel(y,x)], 
				M_src, V_src, M_dst, V_dst, x_dst);

			float *q = warpField.PixelAddress(y,x);
			
			q[0] = float(x_dst[0]);
			q[1] = float(x_dst[1]);

			if (alpha_disp_lbl != INVALID_LABEL)
			{
				twoStepWarp(x_src, lbl2invDepth[alpha_disp_lbl], 
					M_src, V_src, M_dst, V_dst, x_dst);
			
				q[2] = float(x_dst[0]);
				q[3] = float(x_dst[1]);
			}
		}
	}
}


void computeDegFuncArray(
	const Parameters &params,
	const int imageIndex,
	DataIn &dataIn)
{
	const ImageOf<short> img = dataIn._imArray[imageIndex];
	const ImageOf< vector<Coord> > &nbrEdgeSet 
		= dataIn._nbrEdgeArray[imageIndex];

	int height = img.Height();
	int width  = img.Width();

	ImageOf<float> dfArray;
	dfArray.ReAllocate(height, width, 1);

	Coord imSize(width, height);

	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x)
		{
			Coord p(x,y);
			float d = 0.0f;

			vector<Coord> nbr = nbrEdgeSet.Pixel(y,x);
			size_t edge_set_size = nbr.size();

			for (size_t k=0; k < edge_set_size; k++)
			{
				Coord q = nbr[k];

				d += pixelSimilarityUnnormalized(p, q, img, params);
			}

			if(d < 0.001f)
			{
				d = 1.0f;
			}

			dfArray.Pixel(y, x) = d;
		}
	}

	CopyPixels(dataIn._degFuncArray[imageIndex], dfArray);
}


void error_function(
	char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}
