///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef MVSTEREO_H
#define MVSTEREO_H

#include "data.h"


#define VAR_ACTIVE ((Energy::Var)0)
#define VALUE0 0
#define VALUE1 1
#define INVALID_LABEL -1


void multiViewMatch(
	const Parameters       &params,
	const DataIn           &in, 
	DataOut                &out);

void computeImageBound(
	const ImageOf <short>  &img,
	ImageOf <short>        &bd);

void forwardWarpFieldInt(
	const Camera           &srcCam, 
	const ImageOf <short>  &srcLabel,
	const short            alpha,
	const valarray<double> &lbl2invDepth,
	const Camera           &dstCam,
	ImageOf<int>           &warpField);

void forwardWarpFieldFloat(
	const Camera           &srcCam, 
	const ImageOf <short>  &srcLabel,
	const short            alpha,
	const valarray<double> &lbl2invDepth,
	const Camera           &dstCam,
	ImageOf<float>         &warpField);

void computeDegFuncArray(
	const Parameters       &params,
	const int              imageIndex,
	DataIn                 &dataIn);

void error_function(
	char *msg);


////////////////////////////// INLINE FUNCTIONS ///////////////////////////////

inline int round2(
	double x)
{
	return int(floor(x+0.5));
}

inline int intervalDistance(
	const int x, 
	const int x0, 
	const int x1)
{
	if (x < x0) return x0-x;
	else if (x > x1) return x-x1;
	else return int(0);
}

inline int intervalDistance(
	const short x, 
	const short x0, 
	const short x1)
{
	if (x < x0) return x0-x;
	else if (x > x1) return x-x1;
	else return int(0);
}

inline float intervalDistance(
	const float x, 
	const float x0, 
	const float x1)
{
	if (x < x0) return x0-x;
	else if (x > x1) return x-x1;
	else return 0.0f;
}

template<typename T>
inline void boundingColor(
	T       minColor[3], 
	T       maxColor[3], 
	const T color[3])
{
	minColor[0] = min(minColor[0], color[0]);
	minColor[1] = min(minColor[1], color[1]);
	minColor[2] = min(minColor[2], color[2]);

	maxColor[0] = max(maxColor[0], color[0]);
	maxColor[1] = max(maxColor[1], color[1]);
	maxColor[2] = max(maxColor[2], color[2]);
}

template<typename T>
inline void boundingColor(
	T       minColor[3], 
	T       maxColor[3], 
	const T color1[3], 
	const T color2[3])
{
	T tmp;

	tmp = (color1[0] + color2[0])/2;
	minColor[0] = min(minColor[0], tmp);
	maxColor[0] = max(maxColor[0], tmp);

	tmp = (color1[1] + color2[1])/2;
	minColor[1] = min(minColor[1], tmp);
	maxColor[1] = max(maxColor[1], tmp);

	tmp = (color1[2] + color2[2])/2;
	minColor[2] = min(minColor[2], tmp);
	maxColor[2] = max(maxColor[2], tmp);
}

inline void bilinearInterp(
	const Coordf         &p, 
	const ImageOf<short> &img, 
	float                &red, 
	float                &green, 
	float                &blue)
{
	// it is assumed that p is inside the bounds of img

	const int x1 = int(floor(p.x));
	const int x2 = x1 + 1;
	const int y1 = int(floor(p.y));
	const int y2 = y1 + 1;

	const short* tl = img.PixelAddress(y1, x1);
	const short* tr = img.PixelAddress(y1, x2);
	const short* bl = img.PixelAddress(y2, x1);
	const short* br = img.PixelAddress(y2, x2);

	const float w_x1 = x2 - p.x;
	const float w_x2 = p.x - x1;
	const float w_y1 = y2 - p.y;
	const float w_y2 = p.y - y1;

	const float w_tl = w_x1 * w_y1;
	const float w_tr = w_x2 * w_y1;
	const float w_bl = w_x1 * w_y2;
	const float w_br = w_x2 * w_y2;

	red = tl[0]*w_tl + tr[0]*w_tr + bl[0]*w_bl + br[0]*w_br;
	green = tl[1]*w_tl + tr[1]*w_tr + bl[1]*w_bl + br[1]*w_br;
	blue = tl[2]*w_tl + tr[2]*w_tr + bl[2]*w_bl + br[2]*w_br;
}

inline void bilinearInterpBound(
	const Coordf         &p, 
	const ImageOf<short> &imgBound, 
	float                &r_x0, 
	float                &g_x0, 
	float                &b_x0, 
	float                &r_x1, 
	float                &g_x1, 
	float                &b_x1)
{
	// it is assumed that p is inside the bounds of img

	const int x1 = int(floor(p.x));
	const int x2 = x1 + 1;
	const int y1 = int(floor(p.y));
	const int y2 = y1 + 1;

	const short* tl = imgBound.PixelAddress(y1, x1);
	const short* tr = imgBound.PixelAddress(y1, x2);
	const short* bl = imgBound.PixelAddress(y2, x1);
	const short* br = imgBound.PixelAddress(y2, x2);

	const float w_x1 = x2 - p.x;
	const float w_x2 = p.x - x1;
	const float w_y1 = y2 - p.y;
	const float w_y2 = p.y - y1;

	const float w_tl = w_x1 * w_y1;
	const float w_tr = w_x2 * w_y1;
	const float w_bl = w_x1 * w_y2;
	const float w_br = w_x2 * w_y2;

	r_x0 = tl[0]*w_tl + tr[0]*w_tr + bl[0]*w_bl + br[0]*w_br;
	g_x0 = tl[1]*w_tl + tr[1]*w_tr + bl[1]*w_bl + br[1]*w_br;
	b_x0 = tl[2]*w_tl + tr[2]*w_tr + bl[2]*w_bl + br[2]*w_br;
	r_x1 = tl[3]*w_tl + tr[3]*w_tr + bl[3]*w_bl + br[3]*w_br;
	g_x1 = tl[4]*w_tl + tr[4]*w_tr + bl[4]*w_bl + br[4]*w_br;
	b_x1 = tl[5]*w_tl + tr[5]*w_tr + bl[5]*w_bl + br[5]*w_br;
}

inline float pixelSimilarityUnnormalized(
	const Coord          &p, 
	const Coord          &q, 
	const ImageOf<short> &img, 
	const Parameters     &params)
{
	int x_diff = p.x - q.x;
	int y_diff = p.y - q.y;
	float g_dist_sqrd = float(x_diff*x_diff + y_diff*y_diff);
	g_dist_sqrd /= (2.0f * params._sigmaX * params._sigmaX);

	int r_diff = img.Pixel(p.y, p.x, 0) - img.Pixel(q.y, q.x, 0);
	int g_diff = img.Pixel(p.y, p.x, 1) - img.Pixel(q.y, q.x, 1);
	int b_diff = img.Pixel(p.y, p.x, 2) - img.Pixel(q.y, q.x, 2);
	float c_dist_sqrd = float(r_diff*r_diff + g_diff*g_diff + b_diff*b_diff);
	c_dist_sqrd /= (2.0f * params._sigmaC * params._sigmaC);

	return exp(-g_dist_sqrd - c_dist_sqrd);
}

inline float pixelSimilarity(
	const Coord          &p,
	const Coord          &q,
	const ImageOf<float> &degFuncArray,
	const ImageOf<short> &img,
	const Parameters     &params)
{
	float k = pixelSimilarityUnnormalized(p, q, img, params);
	float dp = degFuncArray.Pixel(p.y, p.x);
	float dq = degFuncArray.Pixel(q.y, q.x); 

	// normalize k and return the result
	return k * (1/dp + 1/dq);
}

inline float penaltyPh(
	const Coord          &p, 
	const Coordf         &q, 
	const ImageOf<short> &imgLeft,
	const ImageOf<short> &imgRight,
	const Parameters     &params)
{
	float d_red, d_green, d_blue, d_sum = 0.0f;
	float rr, gr, br;

	const short* Il = imgLeft.PixelAddress(p.y, p.x); 

	float rl = float(Il[0]);
	float gl = float(Il[1]);
	float bl = float(Il[2]);

	bilinearInterp(q, imgRight, rr, gr, br);

	d_red = rl - rr;
	if (params._dataCostType==Parameters::L2) d_red *= d_red;
	else d_red = fabs(d_red);

	d_green = gl - gr;
	if (params._dataCostType==Parameters::L2) d_green *= d_green;
	else d_green = fabs(d_green);

	d_blue = bl - br;
	if (params._dataCostType==Parameters::L2) d_blue *= d_blue;
	else d_blue = fabs(d_blue);

	d_sum = 0.30f * d_red + 0.59f * d_green + 0.11f * d_blue;
	return min(d_sum - params._tauPh,0.0f);
}

inline float penaltyPh(
	const Coord          &p, 
	const Coordf         &q, 
	const ImageOf<short> &imgLeft,
	const ImageOf<short> &imgLeftBound,
	const ImageOf<short> &imgRight,
	const ImageOf<short> &imgRightBound, 
	const Parameters     &params)
{
	float dl, dr  = 0.0f;
	float d_red, d_green, d_blue, d_sum = 0.0f;
	float rr, gr, br;
	float rr_bd_x0, rr_bd_x1, gr_bd_x0, gr_bd_x1, br_bd_x0, br_bd_x1;

	const short* Il = imgLeft.PixelAddress(p.y, p.x); 
	float rl = float(Il[0]);
	float gl = float(Il[1]);
	float bl = float(Il[2]);

	const short* Il_bd = imgLeftBound.PixelAddress(p.y, p.x);
	float rl_bd_x0 = float(Il_bd[0]);
	float gl_bd_x0 = float(Il_bd[1]);
	float bl_bd_x0 = float(Il_bd[2]);
	float rl_bd_x1 = float(Il_bd[3]);
	float gl_bd_x1 = float(Il_bd[4]);
	float bl_bd_x1 = float(Il_bd[5]);

	bilinearInterp(q, imgRight, rr, gr, br);
	bilinearInterpBound(q, imgRightBound, rr_bd_x0, gr_bd_x0, br_bd_x0, rr_bd_x1, gr_bd_x1, br_bd_x1);

	dl = intervalDistance(rl, rr_bd_x0, rr_bd_x1);
	dr = intervalDistance(rr, rl_bd_x0, rl_bd_x1);
	d_red = min(dl, dr); 
	if (params._dataCostType==Parameters::L2) d_red *= d_red;

	dl = intervalDistance(gl, gr_bd_x0, gr_bd_x1);
	dr = intervalDistance(gr, gl_bd_x0, gl_bd_x1);
	d_green = min(dl, dr); 
	if (params._dataCostType==Parameters::L2) d_green *= d_green;

	dl = intervalDistance(bl, br_bd_x0, br_bd_x1);
	dr = intervalDistance(br, bl_bd_x0, bl_bd_x1);
	d_blue = min(dl, dr);  
	if (params._dataCostType==Parameters::L2) d_blue *= d_blue;

	d_sum = 0.30f*d_red + 0.59f*d_green + 0.11f*d_blue;
	return min(d_sum - params._tauPh,0.0f);
}

inline float penaltySm(
   	const Coord          &p, 
	const Coord          &q, 
	const short          lp,
	const short          lq,
	const ImageOf<short> &img,
	const ImageOf<float> &dfArray,
	const Parameters     &params)
{
	int xdiff = p.x-q.x;
	int ydiff = p.y-q.y;
	float dist = sqrtf(float(xdiff*xdiff + ydiff*ydiff));

	float disp_diff = (float)(lp-lq);
	if (disp_diff < 0) disp_diff = -disp_diff;

	disp_diff /= dist;

	if (disp_diff > params._tauSm) disp_diff = params._tauSm;

	float k = pixelSimilarity(p, q, dfArray, img, params);

	return k*disp_diff*params._lambda;
}

inline float penaltyVs(
	const short      lp,
	const short      lq,
	const Parameters &params)
{
	if(lp > lq)
	{
		return params._lambdaVs;
	}
	else
	{
		return 0;
	}
}

inline void twoStepWarp(
	const Vec3double   &x_src,
	const double       &iz,
	const SqMat3double &M_src, 
	const Vec3double   &V_src, 
	const SqMat3double &M_dst,
	const Vec3double   &V_dst,
	Vec3double         &x_dst)
{
	// rotate the source image coordinate back to a reference
	Vec3double tmp = M_src*x_src; 
	Vec3double x_ref;
	x_ref[0] = tmp[0]/tmp[2]*(1+iz*V_src[2])-iz*V_src[0];
	x_ref[1] = tmp[1]/tmp[2]*(1+iz*V_src[2])-iz*V_src[1];
	x_ref[2] = 1;
	x_dst = M_dst*x_ref+iz*V_dst;
	x_dst[0] = x_dst[0]/x_dst[2];
	x_dst[1] = x_dst[1]/x_dst[2];	
}

#endif 


