///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include "pixmap.h"

bool LoadPFM(const char *pathname, float *&pixels, 
			 int &width, int &height, float &maxval, int &nbands)
{
	FILE *fp;
	if ( ( fp = fopen( pathname, "rb" ) ) == NULL )
	{
		printf( "fail to open file %s!\n", pathname );
		return false;
	}	

	//read header

	read_comment(fp);	
	char magic[16];
	fscanf(fp, "%s", magic);
	if (strcmp(magic, "PF")==0)	nbands = 3;
	else if (strcmp(magic, "Pf")==0) nbands = 1;
	else 
	{
		printf("LoadPFM : invalid magic number in %s!\n", pathname);
		return false;
	}
	fgetc(fp);

	read_comment(fp);
	fscanf(fp, "%d", &width);
	if (width <= 0)
	{
		printf("LoadPFM : invalid width in %s!\n", pathname);
		return false;
	}
	fgetc(fp);

	read_comment(fp);
	fscanf(fp, "%d", &height);
	if (height <= 0)
	{
		printf("LoadPFM : invalid height in %s!\n", pathname);
		return false;
	}
	fgetc(fp);

	read_comment(fp);
	fscanf(fp, "%f", &maxval);
	if (maxval == 0) 
	{
		printf("LoadPFM : invalid maxval in %s\n", pathname);
		return false;
	}
	fgetc(fp);

	//read pixels 
	int numPixels = width * height * nbands;
	if (pixels == NULL)
	{
		pixels = new float[numPixels];
	}

	size_t n = fread(pixels, sizeof(float), numPixels, fp);
	fclose(fp);

	if (n == numPixels)
	{
		if (maxval>0) SwapBytes((unsigned char*)pixels, numPixels, sizeof(float)); //big endian 
		else maxval = -maxval; //little endian

		return true;
	}
	else
	{
		printf("LoadPGM : fail to read pixels in %s!\n", pathname);
		return false;
	}
}

bool SavePFM(const char *pathname, const float *pixels, 
			 const int width, const int height, const float maxval, const int nbands)
{
	FILE *fp;
	if ( ( fp = fopen( pathname, "wb" ) ) == NULL )
	{
		printf( "fail to open file %s\n", pathname );
		return false;
	}	

	//write header

	if (nbands == 1) fprintf(fp, "%s\n", "Pf");
	else if (nbands == 3) fprintf(fp, "%s\n", "PF");
	else
	{
		printf("unsupported number of bands for PFM!\n");
		return false;
	}
	fprintf(fp, "%d %d\n", width, height);
	fprintf(fp, "%f\n", -maxval);

	//write pixels 
	int numPixels = width * height * nbands;

	size_t n = fwrite(pixels, sizeof(float), numPixels, fp);
	fclose(fp);

	return (n == numPixels) ? true : false;
}
