///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Copyright (C) 2010 Brandon M. Smith and Li Zhang                          //
//                                                                           //
// Author:       Brandon M. Smith, Dept. of Computer Sciences, UW-Madison    //
// Email:        bmsmith@cs.wisc.edu                                         //
// Personal web: http://www.cs.wisc.edu/~bmsmith                             //
// Project web:  http://www.cs.wisc.edu/~lizhang/projects/mvstereo/cvpr2009/ //
// Description:                                                              //
//                                                                           //
//   This is a C++ implementation of the multi-view stereo matching          //
// algorithm described in:                                                   //
//                                                                           //
//     Brandon M. Smith, Li Zhang, Hailin Jin. Stereo Matching with Non-     //
//     parametric Smoothness Priors in Feature Space. IEEE Computer Society  //
//     Conference on Computer Vision and Pattern Recognition, June 2009.     //
//                                                                           //
// This program is free software: you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by the     //
// Free Software Foundation, either version 3 of the License, or (at your    //
// option) any later version.                                                //
//                                                                           //
//   This program is distributed in the hope that it will be useful, but     //
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-      //
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public  //
// License for more details.                                                 //
//                                                                           //
//   You should have received a copy of the GNU General Public License along //
// with this program.  If not, see <http://www.gnu.org/licenses/>.           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include <valarray>
#include "CmdArgParser.h"
#include "Image.h"
#include "trilatfilt.h"

using std::valarray;


void trilateralFilter(
	const Parameters &params,
	const int        imageIndex,
	const DataIn     &dataIn,
	DataOut          &dataOut)
{
	printf("Running trilateral filter on disparity/depth map %d...\n", imageIndex+1);

	ImageOf<short> img = dataIn._imArray[imageIndex];
	ImageOf<short> idpt = dataOut._lblArray[imageIndex];
	ImageOf<double> odpt;

	const int height = idpt.Height();
	const int width  = idpt.Width();

	odpt.ReAllocate(height, width, 1);

	double orig_dpt, other_dpt;
	double counter, total;
	double Wg, Wc, Wd;
	double dpt_diff, c_diff;
	double weight;
	const int box_radius = params._maxNbrRad;
	const double sigmaXsqrd2 = 2.0 * params._sigmaX * params._sigmaX;
	const double sigmaCsqrd2 = 2.0 * params._sigmaC * params._sigmaC;
	const double sigmaDsqrd2 = 2.0 * params._sigmaD * params._sigmaD;

	for(int y = 0; y < height; y++)
	{
		for(int x = 0; x < width; x++)
		{
			orig_dpt = double(idpt.Pixel(y,x,0));

			counter = 0;
			total   = 0;

			for(int wy = -box_radius; wy <= box_radius; wy++)
			{
				int yp = y + wy;

				if(yp >= 0 && yp < height)
				{
					for(int wx = -box_radius; wx <= box_radius; wx++)
					{
						int xp = x + wx;

						if(xp >= 0 && xp < width)
						{
							other_dpt = double(idpt.Pixel(yp, xp, 0));

							Wg = wx*wx + wy*wy;

							short* p0 = img.PixelAddress(y,x,0);
							short* p1 = img.PixelAddress(yp,xp,0);

							Wc = 0;

							c_diff = double(*p0 - *p1);
							Wc += c_diff*c_diff;
							c_diff = double(*(p0+1) - *(p1+1));
							Wc += c_diff*c_diff;
							c_diff = double(*(p0+2) - *(p1+2));
							Wc += c_diff*c_diff;

							dpt_diff = other_dpt - orig_dpt;
							Wd = dpt_diff*dpt_diff;

							weight = exp(-Wg/sigmaXsqrd2) * exp(-Wc/sigmaCsqrd2) * exp(-Wd/sigmaDsqrd2);

							counter += weight;
							total += weight*other_dpt;
						}
					} // end for(int wx...
				}
			} // end for(int wy...

			odpt.Pixel(y,x,0) = total / counter;

		} // end for(int x...
	} // end for(int y...

	CopyPixels(dataOut._dptSmoothArray[imageIndex], odpt);
}

// Remove disparity/depth map noise
void removeNoise(
	const Parameters &params, 
	const int        imageIndex,
	DataOut          &dataOut)
{
	printf("Running noise removal on disparity/depth map %d...\n", imageIndex+1);

	// if a pixel has fewer than buddy_count_thresh neighboring
	// pixels with the same disparity/depth label, then overwrite
	// it with the mode average surrounding disparity/depth label
	int buddy_count_thresh = params._nrThresh; 

	int numCams   = (int)params._camFiles.size();
	int height    = dataOut._lblArray[imageIndex].Height();
	int width     = dataOut._lblArray[imageIndex].Width();
	Coord imSize(width, height);
	Coord imZero(0, 0);

	vector<Coord> nbr;
	nbr.push_back(Coord(-1,-1));
	nbr.push_back(Coord( 0,-1));
	nbr.push_back(Coord( 1,-1));
	nbr.push_back(Coord(-1, 0));
	nbr.push_back(Coord( 1, 0));
	nbr.push_back(Coord(-1, 1));
	nbr.push_back(Coord( 0, 1));
	nbr.push_back(Coord( 1, 1));

	for(int y = 0; y < height; y++)
	{
		for(int x = 0; x < width; x++)
		{
			Coord p(x, y);
			int lp = dataOut._lblArray[imageIndex].Pixel(y,x);

			int buddy_count = 0;
			vector< Coord > lblvec; //x = label, y = count
			for(int n = 0; n < (int)nbr.size(); n++)
			{
				Coord q = p + nbr[n];

				if(q >= imZero && q < imSize)
				{
					int lq = dataOut._lblArray[imageIndex].Pixel(q.y, q.x); 

					int index = -1;
					for(int k = 0; k < (int)lblvec.size(); k++)
					{
						if(lblvec[k].x == lq)
						{
							index = k;
						}
					}
					if(index < 0)
					{
						lblvec.push_back(Coord(lq, 1));
					}
					else
					{
						lblvec[index].y++;
					}

					if(abs(lp - lq) <= 1)
					{
						buddy_count++;
						if(buddy_count >= buddy_count_thresh)
						{
							break;
						}
					}
				}
			}

			if(buddy_count < buddy_count_thresh)
			{
				int maxCount = 0;
				int modeAvgLbl = 0;
				for(int k = 0; k < (int)lblvec.size(); k++)
				{
					if(lblvec[k].y > maxCount && lblvec[k].x > modeAvgLbl)
					{
						maxCount = lblvec[k].y;
						modeAvgLbl = lblvec[k].x;
					}
				}
				dataOut._lblArray[imageIndex].Pixel(y, x) = modeAvgLbl; 
			}
		}
	}
}
